//
// shatbox (C) 2002 Matthew Conte (matt@conte.com)
//
//
// apu.cpp
//
// apu emulation
//
// $Id: $
//

#include <string.h>
#include "assert.h"
#include "regs.h"
#include "apu.h"
#include "cpu.h"

// length counter decoded values
static const uint8 lengthLookup[32] = 
{
    5, 127,
   10,   1,
   19,   2,
   40,   3,
   80,   4,
   30,   5,
    7,   6,
   13,   7,
    6,   8,
   12,   9,
   24,  10,
   48,  11,
   96,  12,
   36,  13,
    8,  14,
   16,  15
};


// base class
void soundChannel::reset()
{
   mOutput = 0;
   mOutputTimer = 0;
   mTimer = 0;
   mFreq = 1; // prevent divide by zero
   mLengthCounter = 0;
   
   for (int i = 0; i < 4; i++)
      mRegs[i] = 0;
   
   mEnabled = false;
}

void soundChannel::write(int reg, uint8 value)
{
   ASSERT(reg >= 0 && reg < 4);
   mRegs[reg] = value;
}

void soundChannel::phaseChange(int /*phase*/)
{
}

void soundChannel::activate(bool state)
{
   mEnabled = state;
   if (!state)
      mLengthCounter = 0;
}

bool soundChannel::isActive()
{
   if (mLengthCounter)
      return true;

   return false;
}



// rectangle
rectangleChannel::rectangleChannel(int which)
: mWhich(which)
{
   reset();
}

void rectangleChannel::reset()
{
   soundChannel::reset();

   mStep = 0;
   mEnvCount = 0;
   mEnvRate = 1; // prevent divide by zero
   mSweepRate = 1;
   mSweepCarry = 1; // TODO: find initial value
   mDutyPeriod = 1; // TODO: find initial value
}

void rectangleChannel::write(int reg, uint8 value)
{
   // fractional length (/16) of positive to negative rect wave amplitude
   const int dutyPeriod[4] = 
   { 
      2, 4, 8, 12 
   };

   // values of rect freqs which in increase mode will overflow 11 bits
   const uint16 sweepCarryLookup[8] =
   {
      0x400, 0x556, 0x667, 0x71D, 
      0x788, 0x7C2, 0x7E1, 0x7F1
   };

   soundChannel::write(reg, value);

   switch (reg)
   {
   case 0:
      mEnvRate = (value & REGF_A0_ENVMASK) + 1;
      mDutyPeriod = dutyPeriod[value >> REGF_A0_DUTYSHIFT];
      break;

   case 1:
      mSweepRate = ((value & REGF_A1_SLEN) >> 4) + 1;
      mSweepCarry = sweepCarryLookup[value & REGF_A1_SSHIFT];
      break;

   case 2:
      mFreq = ((mRegs[3] & 7) << 8) + value + 1;
      break;

   case 3:
      mEnvCount = 0x0f;
      mStep = 0;
      mLengthCounter = lengthLookup[value >> 3];
      mFreq = ((value & 7) << 8) + mRegs[2] + 1;
      break;
   }
}

int rectangleChannel::getSample(int rate)
{
   mOutputTimer -= rate;
   while (mOutputTimer < 0)
   {
      mOutputTimer += apu::APU_FRAMELEN / DECAY_DIV;
      mOutput -= mOutput >> DECAY_SHIFT;
   }

   if (mEnabled && mLengthCounter && mFreq >= 8
       && !((0 == (mRegs[1] & REGF_A1_SDEC) && mFreq >= mSweepCarry)))
   {
      mTimer -= rate;
      while (mTimer < 0)
      {
         ASSERT(mFreq != 0);
         mTimer += mFreq << apu::APU_FIX;
         mStep = (mStep + 1) & 0x0f;

         if (0 == mStep)
         {
            if (mRegs[0] & REGF_A0_ENVFIX)
               mOutput += (mRegs[0] & REGF_A0_ENVMASK) << apu::OUTPUT_FIX;
            else
               mOutput += mEnvCount << apu::OUTPUT_FIX;
         }
         else if (mStep == mDutyPeriod)
         {
            if (mRegs[0] & REGF_A0_ENVFIX)
               mOutput -= (mRegs[0] & REGF_A0_ENVMASK) << apu::OUTPUT_FIX;
            else
               mOutput -= mEnvCount << apu::OUTPUT_FIX;
         }
      }
   }

   return mOutput;
}

void rectangleChannel::phaseChange(int phase)
{
   ASSERT(mEnvRate);
   if (--mEnvRate == 0)
   {
      mEnvRate = (mRegs[0] & REGF_A0_ENVMASK) + 1;

      if (mRegs[0] & REGF_A0_LOOPED)
         mEnvCount = (mEnvCount - 1) & 0x0f;
      else if (mEnvCount > 0)
         mEnvCount--;
   }

   if (phase & 1)
   {
      ASSERT(mSweepRate);
      if ((mRegs[1] & REGF_A1_SENABLE) && --mSweepRate == 0)
      {
         mSweepRate = ((mRegs[1] & REGF_A1_SLEN) >> 4) + 1;
         if (mRegs[1] & REGF_A1_SSHIFT)
         {         
            if (mRegs[1] & REGF_A1_SDEC)
            {
               if (0 == mWhich)
                  mFreq -= (mFreq >> (mRegs[1] & REGF_A1_SSHIFT)) + 1;
               else
                  mFreq -= mFreq >> (mRegs[1] & REGF_A1_SSHIFT);
            }
            else
            {
               mFreq += mFreq >> (mRegs[1] & REGF_A1_SSHIFT);
            }
         }
      }
   }

   if (3 == (phase & 3))
   {
      if (0 == (mRegs[0] & REGF_A0_LOOPED) && mLengthCounter)
         mLengthCounter--;
   }
}


// triangle
triangleChannel::triangleChannel()
: soundChannel()
{
   reset();
}

void triangleChannel::reset()
{
   soundChannel::reset();
   mStep = 0;
   mLatency = 0;
   mCounting = false;
   mLinearCounter = 0;
}

void triangleChannel::write(int reg, uint8 value)
{
   soundChannel::write(reg, value);

   switch (reg)
   {
   case 0:
      if (false == mCounting && mLengthCounter)
         mLinearCounter = value & REGF_C0_LENMASK;
      break;

   case 2:
      mFreq = ((mRegs[3] & 7) << 8) + value + 1;
      break;

   case 3:
      // one frame latency
      mLatency = apu::APU_FRAMELEN;
      mCounting = false;

      mLinearCounter = mRegs[0] & REGF_C0_LENMASK;

      mLengthCounter = lengthLookup[value >> 3];
      mFreq = ((value & 7) << 8) + mRegs[2] + 1;
      break;

   default:
      break;
   }
}

int triangleChannel::getSample(int rate)
{
   mOutputTimer -= rate;
   while (mOutputTimer < 0)
   {
      mOutputTimer += apu::APU_FRAMELEN / DECAY_DIV;
      mOutput -= mOutput >> DECAY_SHIFT;
   }

   if (!mCounting && (0 == (mRegs[0] & REGF_C0_LOOPED))
       && mLatency)
   {
      mLatency -= rate;
      if (mLatency < 0)
      {
         mLatency = 0;
         mCounting = true;
      }
   }

   if (mEnabled && mLengthCounter && mLinearCounter && mFreq >= 4)
   {
      mTimer -= rate;
      while (mTimer < 0)
      {
         ASSERT(mFreq != 0);

         mTimer += mFreq << apu::APU_FIX;
         mStep = (mStep + 1) & 0x1f;

         if (mStep & 0x10)
            mOutput += 1 << apu::OUTPUT_FIX;
         else
            mOutput -= 1 << apu::OUTPUT_FIX;
      }
   }

   return mOutput;
}

void triangleChannel::phaseChange(int phase)
{
   if (mCounting && mLinearCounter)
      mLinearCounter--;

   if (3 == (phase & 3))
   {
      if (0 == (mRegs[0] & REGF_C0_LOOPED) && mLengthCounter)
         mLengthCounter--;
   }
}


// noise
noiseChannel::noiseChannel()
: soundChannel()
{
   reset();
}

void noiseChannel::reset()
{
   soundChannel::reset();
   mEnvCount = 0;
   mEnvRate = 1;
   mState = 0;
   mShiftReg = 1 << 14;
   mXorTap = 1 << 1;
}

uint8 noiseChannel::shiftRegister15()
{
   int bit0 = mShiftReg & 1;
   int tap = (mShiftReg & mXorTap) ? 1 : 0;
   int bit14 = (bit0 ^ tap) << 14;
   mShiftReg >>= 1;
   mShiftReg |= bit14;
   
   return (bit0 ^ 1);
}

void noiseChannel::write(int reg, uint8 value)
{
   // noise channel 11-bit wavelength converter * 2
   const uint16 noiseLookup[16] =
   {
        4,    8,   16,   32,
       64,   96,  128,  160,
      202,  254,  380,  508,
      762, 1016, 2034, 4068
   };

   soundChannel::write(reg, value);

   switch (reg)
   {
   case 0:
      mEnvRate = (value & REGF_D0_ENVMASK) + 1;
      break;

   case 2:
      mFreq = noiseLookup[value & REGF_D2_FREQMASK];
      mXorTap = (value & REGF_D2_MODE) ? (1 << 6) : (1 << 1);
      break;

   case 3:
      mEnvCount = 0x0f;
      mLengthCounter = lengthLookup[value >> REGF_D3_LENRSHIFT];
      break;

   default:
      break;
   }
}

int noiseChannel::getSample(int rate)
{
   mOutputTimer -= rate;
   while (mOutputTimer < 0)
   {
      mOutputTimer += apu::APU_FRAMELEN / DECAY_DIV;
      mOutput -= mOutput >> DECAY_SHIFT;
   }

   if (mEnabled && mLengthCounter)
   {
      int vol;

      if (mRegs[0] & REGF_D0_ENVFIX)
         vol = mRegs[0] & REGF_D0_ENVMASK;
      else
         vol = mEnvCount;

      mTimer -= rate;
      while (mTimer < 0)
      {
         ASSERT(mFreq != 0);
         mTimer += mFreq << apu::APU_FIX;

         uint8 state = shiftRegister15();
         if (mState != state)
         {
            mState = state;
            if (state)
               mOutput += vol << apu::OUTPUT_FIX;
            else
               mOutput -= vol << apu::OUTPUT_FIX;
         }
      }
   }

   return mOutput;
}

void noiseChannel::phaseChange(int phase)
{
   ASSERT(mEnvRate);
   if (--mEnvRate == 0)
   {
      mEnvRate = (mRegs[0] & REGF_D0_ENVMASK) + 1;

      if (mRegs[0] & REGF_D0_LOOPED)
         mEnvCount = (mEnvCount - 1) & 0x0f;
      else if (mEnvCount > 0)
         mEnvCount--;
   }

   if (3 == (phase & 3))
   {
      if (0 == (mRegs[0] & REGF_D0_LOOPED) && mLengthCounter)
         mLengthCounter--;
   }
}


// delta modulation channel
dmcChannel::dmcChannel(cpu *pCPU)
: soundChannel(),
  mpCPU(pCPU)
{
   reset();
}

void dmcChannel::reset()
{
   soundChannel::reset();
   mLength = 0;
   mShiftReg = 0;
   mAddress = 0;
   mIRQFlag = false;
   mOldVol = 0;
}

void dmcChannel::write(int reg, uint8 value)
{
   // dmc transfer clock table (synthesized per bit)
   const uint16 dmcLookup[16] = 
   {
      0x1ac, 0x17c, 0x154, 0x140,
      0x11e, 0x0fe, 0x0e2, 0x0d6,
      0x0be, 0x0a0, 0x08e, 0x080,
      0x06a, 0x054, 0x048, 0x036
   };

   soundChannel::write(reg, value);

   switch (reg)
   {
   case 0:
      if (0 == (value & 0x80))
         mIRQFlag = false; // clear irq flag
      mFreq = dmcLookup[value & REGF_E0_FREQMASK];
      break;

   case 1:
      value &= 0x7f; // MSB unused
      mRegs[1] = value;
      mOutput += (value - mOldVol) << apu::OUTPUT_FIX;
      mOldVol = value;
      break;

   default:
      break;
   }
}

int dmcChannel::getSample(int rate)
{
   mOutputTimer -= rate;
   while (mOutputTimer < 0)
   {
      mOutputTimer += apu::APU_FRAMELEN / DECAY_DIV;
      mOutput -= mOutput >> DECAY_SHIFT;
   }

   if (mEnabled && mLength)
   {
      mTimer -= rate;
      while (mTimer < 0)
      {
         ASSERT(mFreq != 0);
         mTimer += mFreq << apu::APU_FIX;

         // reload a byte into shift register
         if (0 == (mLength & 7)) // check bits
         {
            mShiftReg = mpCPU->dma(mAddress);
            mAddress = 0x8000 | ((mAddress + 1) & 0x7fff);
         }

         int bit = mShiftReg & 1;
         mShiftReg >>= 1;
         mLength--;
         if (0 == mLength)
         {
            // handle irq / reload
            if (mRegs[0] & REGF_E0_LOOPED)
            {
               activate(true);
            }
            else
            {
               if (mRegs[0] & REGF_E0_IRQGEN)
               {
                  mIRQFlag = true;
                  mpCPU->irq();
               }

               break;
            }
         }

         // adding/subtracting on DMC affects bits 6-1 of
         // REG_E1, as the 6-bit delta counter is tied to
         // bits 6-1 of the 7-bit DAC.
         if (bit)
         {
            if (mRegs[1] < 0x7e)
            {
               mRegs[1] += 2;
               mOutput += 2 << apu::OUTPUT_FIX;
            }
         }
         else
         {
            if (mRegs[1] > 1)
            {
               mRegs[1] -= 2;
               mOutput -= 2 << apu::OUTPUT_FIX;
            }
         }
      }
   }

   return mOutput >> 1;
}

void dmcChannel::activate(bool state)
{
   mIRQFlag = false;

   mEnabled = state;
   if (state)
   {
      if (0 == mLength)
      {
         mAddress = 0xc000 + mRegs[2] * 64;
         mLength = (mRegs[3] * 16 + 1) * 8; // we count bits
      }
   }
   else
   {
      mLength = 0;
   }
}

bool dmcChannel::isActive()
{
   if (mLength)
      return true;

   return false;
}

static const nes6502_memwrite apu_memwrite[] =
{
   { 0x4000, 0x4013, apu::dispatchWrite, NULL },
   { 0x4015, 0x4015, apu::dispatchWriteCtrl, NULL },
   { (uint32) -1, (uint32) -1, NULL, NULL }
};

static const nes6502_memread apu_memread[] =
{
   { 0x4015, 0x4015, apu::dispatchReadCtrl, NULL },
   { (uint32) -1, (uint32) -1, NULL, NULL }
};

// apu
apu::apu(machine *pMachine, cpu *pCPU, int mixingRate)
: unit(pMachine)
{
   mpMemWrite = apu_memwrite;
   mpMemRead = apu_memread;
   
   if (!mixingRate)
      exit(-1);

   mRate = APU_RATE / mixingRate;
   mCycles = 0;
   mPhaseCount = APU_FRAMELEN / 4;
   mPhase = 0;
   mControlReg = 0;

   mpChannel[CHAN_RECT0] = new rectangleChannel(CHAN_RECT0);
   mpChannel[CHAN_RECT1] = new rectangleChannel(CHAN_RECT1);
   mpChannel[CHAN_TRI] = new triangleChannel;
   mpChannel[CHAN_NOISE] = new noiseChannel;
   mpChannel[CHAN_DMC] = new dmcChannel(pCPU);

   mBufferSize = mixingRate / 60;
   mpMixBuffer = new uint8[mBufferSize];
   memset(mpMixBuffer, 0x80, mBufferSize);
   mBufferPos = 0;
}

apu::~apu()
{
   delete [] mpMixBuffer;

   for (int i = 0; i < NUM_CHANNELS; i++)
   {
      delete mpChannel[i];
      mpChannel[i] = NULL;
   }
}

void apu::reset(bool /*hardReset */)
{
   mCycles = 0;
   mPhaseCount = APU_FRAMELEN / 4;
   mPhase = 0;

   for (int i = 0; i < NUM_CHANNELS; i++)
      mpChannel[i]->reset();

   mControlReg = 0; // all channels disabled
}

void apu::update()
{
   int available = mBufferSize - mBufferPos;
   for (; mCycles < 0 && available > 0; mCycles += mRate, available--)
   {
      mPhaseCount -= mRate;
      if (mPhaseCount < 0)
      {
         mPhaseCount += APU_FRAMELEN / 4;
         for (int i = 0; i < NUM_CHANNELS; i++)
            mpChannel[i]->phaseChange(mPhase);
         mPhase = (mPhase + 1) % 4;
      }

      int accum = 0;

      for (int i = 0; i < NUM_CHANNELS; i++)
         accum += mpChannel[i]->getSample(mRate);

      accum >>= (OUTPUT_FIX - 2);

      if (accum < -128)
         accum = -128;
      else if (accum > 127)
         accum = 127;

      mpMixBuffer[mBufferPos++] = accum ^ 0x80;
   }
}

int apu::flushSamples(unsigned char **ppStream)
{
   int size = mBufferPos;
   mBufferPos = 0;
   *ppStream = mpMixBuffer;
   return size;
}

uint8 apu::dispatchReadCtrl(void *userdata, uint32 offset)
{
   return ((apu *) userdata)->readCtrl(offset);
}

void apu::dispatchWrite(void *userdata, uint32 offset, uint8 value)
{
   ((apu *) userdata)->write(offset, value);
}

void apu::dispatchWriteCtrl(void *userdata, uint32 offset, uint8 value)
{
   ((apu *) userdata)->writeCtrl(offset, value);
}

uint8 apu::readCtrl(uint32 /*offset*/)
{
   update();

   uint8 value = 0;

   for (int i = 0; i < NUM_CHANNELS; i++)
   {
      if ((mControlReg & (1 << i)) && mpChannel[i]->isActive())
         value |= (1 << i);
   }

   if (((dmcChannel *) mpChannel[CHAN_DMC])->getIRQFlag())
      value |= REGF_ST_DMCIRQ;

   return value;
}

void apu::write(uint32 offset, uint8 value)
{
   update();

   int channel = offset / 4;
   int reg = offset % 4;

   mpChannel[channel]->write(reg, value);
}

void apu::writeCtrl(uint32 /*offset*/, uint8 value)
{
   update();

   mControlReg = value;

   for (int i = 0; i < NUM_CHANNELS; i++)
      mpChannel[i]->activate((value & (1 << i)) ? true : false);
}
