//
// shatbox (C) 2002 Matthew Conte (matt@conte.com)
//
//
// cpu.cpp
//
// cpu unit
//
// $Id: $
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cpu.h"
#include "dis6502.h"
#include "apu.h"
#include "ppu.h"
#include "input.h"
#include "mmc.h"
#include "nes.h"

// TODO: frame IRQ


#define  CPU_RAM_SIZE         0x800
#define  NES_CLOCK_DIVIDER    12
#define  NES_SCANLINE_CYCLES  (1364 / (float) NES_CLOCK_DIVIDER)


static void cpu_jam(nes6502_context *pContext)
{
   fprintf(stderr, "execution halted due to cpu jam\n");
   fprintf(stderr, nes6502_disasm(pContext->pc_reg - 1, pContext->p_reg,
                                  pContext->a_reg, pContext->x_reg,
                                  pContext->y_reg, pContext->s_reg));
   exit(-1);
}


static uint8 cpuReadProtect(void */*userdata*/, uint32 /*offset*/)
{
   return 0xff;
}

static void cpuWriteProtect(void */*userdata*/, uint32 /*offset*/, uint8 /*offset*/)
{
}

// traps for 'bad' reads/writes
static const nes6502_memread badReadTrap[] = 
{
   { 0x4000, 0x5fff, cpuReadProtect, NULL },
   { (uint32) -1, (uint32) -1, NULL, NULL }
};

static const nes6502_memwrite badWriteTrap[] = 
{
   { 0x4000, 0x5fff, cpuWriteProtect, NULL },
   { 0x8000, 0xffff, cpuWriteProtect, NULL },
   { (uint32) -1, (uint32) -1, NULL, NULL }
};

static const nes6502_memread nullRead =
{
   (uint32) -1, (uint32) -1, NULL, NULL
};

static const nes6502_memwrite nullWrite =
{
   (uint32) -1, (uint32) -1, NULL, NULL
};


cpu::cpu(machine *pMachine)
: unit(pMachine)
{
   nes6502_context context;

   memset(&context, 0, sizeof(nes6502_context));

   for (int i = 0; i < MAX_MEM_HANDLER + 1; i++)
   {
      mReadTrap[i] = nullRead;
      mWriteTrap[i] = nullWrite;
   }

   // allocate 2kB system RAM
   mpRAM = new uint8[CPU_RAM_SIZE];
   if (NULL == mpRAM)
   {
      exit(-1);
   }

   context.mem_page[0] = mpRAM;

   context.jam_callback = cpu_jam;
   context.read_handler = mReadTrap;
   context.write_handler = mWriteTrap;

   nes6502_setcontext(&context);
}

cpu::~cpu()
{
   delete [] mpRAM;
}

void cpu::reset(bool hardReset)
{
   if (hardReset)
   {
      memset(mpRAM, 0, CPU_RAM_SIZE);
   }

   nes6502_reset();

   mCurrentCycles = NES_SCANLINE_CYCLES;
}


int cpu::addReadTrap(const nes6502_memread *pReadTrap, unit *pUnit)
{
   int i;

   if (NULL == pReadTrap)
      return 0;

   for (i = 0; i < MAX_MEM_HANDLER; i++)
   {
      if (mReadTrap[i].min_range == (uint32) -1)
         break;
   }

   do
   {
      if (MAX_MEM_HANDLER == i)
         return -1; // out of slots

      mReadTrap[i] = *pReadTrap;
      mReadTrap[i].userdata = (void *) pUnit;
      pReadTrap++;
      i++;
   } 
   while (pReadTrap->min_range != (uint32) -1);

   mReadTrap[i] = nullRead;

   return 0;
}

int cpu::addWriteTrap(const nes6502_memwrite *pWriteTrap, unit *pUnit)
{
   int i;

   if (NULL == pWriteTrap)
      return 0;

   for (i = 0; i < MAX_MEM_HANDLER; i++)
      if (mWriteTrap[i].min_range == (uint32) -1)
         break;

   do
   {
      if (MAX_MEM_HANDLER == i)
         return -1; // out of slots

      mWriteTrap[i] = *pWriteTrap;
      mWriteTrap[i].userdata = (void *) pUnit;
      pWriteTrap++;
      i++;
   } 
   while (pWriteTrap->min_range != (uint32) -1);

   mWriteTrap[i] = nullWrite;

   return 0;
}


void cpu::addTraps(unit *pUnit)
{
   addReadTrap(pUnit->getMemRead(), pUnit);
   addWriteTrap(pUnit->getMemWrite(), pUnit);
}

void cpu::closeTraps()
{
   addReadTrap(badReadTrap, (unit *) NULL);
   addWriteTrap(badWriteTrap, (unit *) NULL);
}


/*
; - Any write to $4017 seems to reset the frame counter.
; - bit6 of $4015 will be set when frame counter overflow.
; - Reading from $4015 will clear bit6, and the "frame IRQ" request
;   will be cleared, but does not seem to reset the frame counter.
; - Writing $00 (i.e. bit6=0) to $4015 will NOT reset bit6.
; - The "frame IRQ" will be kept requested until the bit6 of $4015
;   is cleared, thus the IRQ will be triggered immediately
;   after I-flag (of CPU) is cleared.
; - Read from $4017 will always return a value with bit6=1, and
;   bit7=0, despite what is previously written to bit6 and bit7.
*/

void cpu::setFrameIRQ(uint8 value)
{
   if (value & 0x40)
   {
      nes6502_clearirq();
   }
}

void cpu::nmi()
{
   nes6502_nmi();
}

void cpu::irq()
{
   nes6502_irq();
}

uint8 cpu::dma(uint32 address)
{
   uint8 data = nes6502_getbyte(address);
   nes6502_burn(1);
   return data;
}

int cpu::runCycles(int cycles)
{
   int elapsed = nes6502_execute(cycles);
   ((NES *) mpMachine)->getAPU()->advance(elapsed);
   return elapsed;
}

void cpu::vbl()
{
   mCurrentCycles -= (float) runCycles(7);
}

void cpu::scanline()
{
   if (mCurrentCycles > 0)
      mCurrentCycles -= runCycles((int) mCurrentCycles);

   mCurrentCycles += NES_SCANLINE_CYCLES;
}
