;; nes6502.asm
;;
;; Custom NES 6502 (N2A03) CPU emulator
;;
;; Code by matt@conte.com and rvu@idecnet.com
;;
;; $Id: $
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   BITS  32

%macro cglobal 1 
          global _%1 
%define %1 _%1 
%endmacro


%macro cextern 1 
          extern _%1 
%define %1 _%1 
%endmacro

;%define  NES6502_DECIMAL
;%define  NES6502_DISASM
%define  NES6502_UNDOCUMENTED

%define  NES6502_NUMBANKS     16
%define  NES6502_BANKSHIFT    12
%define  NES6502_BANKSIZE     (0x10000 / NES6502_NUMBANKS)
%define  NES6502_BANKMASK     (NES6502_BANKSIZE - 1)

%define  N_FLAG         0x80
%define  V_FLAG         0x40
%define  R_FLAG         0x20
%define  B_FLAG         0x10
%define  D_FLAG         0x08
%define  I_FLAG         0x04
%define  Z_FLAG         0x02
%define  C_FLAG         0x01

%define  NMI_VECTOR     0xFFFA
%define  RESET_VECTOR   0xFFFC
%define  IRQ_VECTOR     0xFFFE

%define  INT_CYCLES     7
%define  RESET_CYCLES   6

%define  STACK_OFFSET   0x0100

%define  NMI_MASK       0x01
%define  IRQ_MASK       0x02

; these soon to be exchanged with al/ah
%define  REG_A       dl
%define  REG_P       dh

%define  REG_X       cl
%define  REG_Y       ch
%define  REG_PC      esi
%define  REG_S       edi

   SECTION  .data

; this structure must match the one in nes6502.h exactly!
   ALIGN 4
nes6502_context:
   mem_page       TIMES NES6502_NUMBANKS  dd 0

   jam_callback   dd 0

   read_handler   dd 0  ; pointer to array of structs (size 4 dwords)
   write_handler  dd 0  ; pointer to array of structs (size 4 dwords)
   
   total_cycles   dd 0
   burn_cycles    dd 0
   remaining_cycles dd 0

   pc_reg         dd 0

   a_reg          db 0
   p_reg          db 0
   x_reg          db 0
   y_reg          db 0

   s_reg          db 0
   jammed         db 0
   int_pending    db 0
   __padding      db 0

nes6502_context_end:

   ALIGN 4
; TODO: should be in text section
; TODO: remove this, using lahf/sahf
; precomputed flag table for fast Z,N setting
flag_table:
               db Z_FLAG
   times 0x7f  db 0
   times 0x80  db N_FLAG

; TODO: dead page
;dead_page:     TIMES NES6502_BANKSIZE   db 0


;-----------------------------------------------------------------------
; macros
;-----------------------------------------------------------------------
; misc macros

%macro ADD_CYCLES 1
%if %1 = 1
   dec   dword [remaining_cycles]
   inc   dword [total_cycles]
%else
   sub   dword [remaining_cycles],%1
   add   dword [total_cycles],%1
%endif
%endmacro


;-----------------------------------------------------------------------
; addressing mode macros
; all read macros leave read in al/ax, and address in ebx

%macro bank_readbyte 1
%ifnidni %1,eax            ; this means, if param1<>eax
   mov   eax,%1            ; i love nasm, we get 1 cycles less if we use
%endif                     ; eax as input
   mov   ebx,eax
   and   eax,NES6502_BANKMASK     ; this makes eax<=0x1fff
                           ; so we don't need "and 0x00001fff" later
   shr   ebx,NES6502_BANKSHIFT
   mov   ebx,[mem_page+ebx*4]   ; get bank offset
   movzx eax,byte [ebx + eax]
%endmacro

%macro bank_readword 1
%ifnidni %1,eax            ; this means, if param1<>eax
   mov   eax,%1            ; i love nasm, we get 1 cycles less if we use
%endif                     ; eax as input
   mov   ebx,eax
   and   eax,NES6502_BANKMASK     ; this makes eax<=0x1fff
   shr   ebx,NES6502_BANKSHIFT
   mov   ebx,[mem_page+ebx*4]   ; get bank offset
   movzx eax,word [ebx+eax]
%endmacro

; write a byte
; warning destroy ebx
; al=value
; ebx=address
%macro bank_writebyte 2
   push  ebp
   mov   ebp,%1
   and   %1,NES6502_BANKMASK
   shr   ebp,NES6502_BANKSHIFT
   mov   ebp,[mem_page+ebp*4]     ;get bank offset
   mov   [ebp+%1],%2        ;put byte inside bank
   pop   ebp
%endmacro

%macro ZP_READBYTE 1
   and   %1,0x000000ff
   mov   al,[ebp + %1]
%endmacro

%macro ZP_READWORD 1
   and   %1,0x000000ff
   movsx eax,word [ebp + %1]
%endmacro

%macro ZP_WRITEBYTE 2
   and   %1,0x000000ff
   mov   [ebp + %1],%2
%endmacro

%macro pushall 0
   push  ebx
   push  ecx
   push  edx
   push  edi
   push  esi
   push  ebp
%endmacro

%macro popall 0
   pop   ebp
   pop   esi
   pop   edi
   pop   edx
   pop   ecx
   pop   ebx
%endmacro

%macro mem_readbyte 0
   ;ram?
   and   eax,0x0000ffff
   cmp   eax,0x800
   jae   %%mem_read_not_ram
   mov   al,[ebp + eax]
   jmp   SHORT %%done

%%mem_read_not_ram:
   ; always paged memory
   cmp   eax,0x8000
   jae   %%end_mem_read

   mov   ebx,[read_handler]   ; get the handler

%%mem_read_rh:
   cmp   dword [ebx],0xffffffff
   je    %%end_mem_read
   cmp   eax,[ebx]         ; min_range
   jb    %%next_mem_read
   cmp   eax,[ebx + 4]     ; max_range
   ja    %%next_mem_read
   ; must call read_func

   pushall
   sub   eax,[ebx]         ; get to beginning of range

   push  eax
   mov   eax,[ebx+12]      ; userdata
   push  eax
   call  [ebx+8]           ; read_func
   add   esp,8

   popall
   jmp   SHORT %%done

%%next_mem_read:
   add   ebx,4*4           ; struct is 4 dwords
   jmp   SHORT %%mem_read_rh

%%end_mem_read:
   ; return paged memory
   bank_readbyte eax

%%done:
%endmacro


%macro mem_writebyte 0
   ;ram?
   and   ebx,0x0000ffff
   cmp   ebx,0x800
   jae   %%mem_write_not_ram
   mov   [ebp + ebx],al
   jmp   SHORT %%done_nopop

%%mem_write_not_ram:
   pushall
   push  eax      ; save on stack for later!
   mov   eax,[write_handler]
%%mem_write_wh:
   cmp   dword [eax],0xffffffff
   je    %%end_mem_write
   cmp   ebx,[eax]        ; min_range
   jb    %%next_mem_write
   cmp   ebx,[eax+4]      ; max_range
   ja    %%next_mem_write

   ; our second param is already on stack!
   sub   ebx,[eax]         ; get to beginning of mem
   push  ebx               ; push first param
   mov   ebx,[eax+12]      ; userdata
   push  ebx
   call  [eax+8]          ; write_func
   add   esp,12
   jmp   SHORT %%done

%%next_mem_write:
   add   eax,4*4           ; struct is 4 dwords
   jmp   SHORT %%mem_write_wh

%%end_mem_write:
   ; write to paged memory
   pop   eax
   bank_writebyte ebx,al      ; eax=value, ebx=address

%%done:
   popall
%%done_nopop:
%endmacro



%macro MEMORY_WRITEBYTE 2
%ifnidni %1,ebx
   mov   ebx,%1
%endif
%ifnidni %2,al
   mov   al,%2
%endif

   mem_writebyte
%endmacro

;empty
%macro EMPTY_READ 0
%endmacro

%macro IMMEDIATE_BYTE 0
   bank_readbyte REG_PC
   inc   REG_PC
%endmacro

%macro ABSOLUTE_ADDR 0
   bank_readword REG_PC
   add   REG_PC,2
%endmacro

%macro ABSOLUTE 0
   ABSOLUTE_ADDR

   ; temp
   push  eax
   mem_readbyte
   pop   ebx
%endmacro

%macro ABSOLUTE_BYTE 0
   ABSOLUTE
%endmacro

%macro ABS_IND_X_ADDR 0
   ABSOLUTE_ADDR
   add   al,REG_X
   jnc   %%nocarry
   inc   ah
   ADD_CYCLES 1
%%nocarry:
%endmacro

%macro ABS_IND_X 0
   ABS_IND_X_ADDR

   ; temp
   push  eax
   mem_readbyte
   pop   ebx
%endmacro

%macro ABS_IND_X_BYTE 0
   ABS_IND_X
%endmacro

%macro ABS_IND_Y_ADDR 0
   ABSOLUTE_ADDR
   add   al,REG_Y
   jnc   %%nocarry
   inc   ah
   ADD_CYCLES 1
%%nocarry:
%endmacro

%macro ABS_IND_Y 0
   ABS_IND_Y_ADDR

   ; temp
   push  eax
   mem_readbyte
   pop   ebx
%endmacro

%macro ABS_IND_Y_BYTE 0
   ABS_IND_Y
%endmacro

%macro ZERO_PAGE_ADDR 0
   IMMEDIATE_BYTE
%endmacro

%macro ZERO_PAGE 0
   ZERO_PAGE_ADDR

   ; TODO temp...
   mov   ebx,eax
   ZP_READBYTE eax
%endmacro

%macro ZERO_PAGE_BYTE 0
   ZERO_PAGE
%endmacro

%macro ZP_IND_X_ADDR 0
   ZERO_PAGE_ADDR
   add   al,REG_X
%endmacro

%macro ZP_IND_X 0
   ZP_IND_X_ADDR

   ; TODO temp...
   mov   ebx,eax
   ZP_READBYTE eax
%endmacro

%macro ZP_IND_X_BYTE 0
   ZP_IND_X
%endmacro

; Zero-page indexed Y doesn't really exist, just for LDX / STX
%macro ZP_IND_Y_ADDR 0
   ZERO_PAGE_ADDR
   add   al,REG_Y
%endmacro

%macro ZP_IND_Y_BYTE 0
   ZP_IND_Y_ADDR

   ; TODO temp...
   mov   ebx,eax
   ZP_READBYTE eax
%endmacro

; Indirect indexed 
%macro INDIR_X_ADDR 0
   ZERO_PAGE_ADDR
   add   al,REG_X
   ZP_READWORD eax
%endmacro

%macro INDIR_X 0
   INDIR_X_ADDR

   ; temp
   push  eax
   mem_readbyte
   pop   ebx
%endmacro

%macro INDIR_X_BYTE 0
   INDIR_X
%endmacro

; Indexed indirect
%macro INDIR_Y_ADDR 0
   ZERO_PAGE_ADDR
   ZP_READWORD eax
   add   al,REG_Y
   jnc   %%nocarry
   inc   ah
   ADD_CYCLES 1
%%nocarry:
%endmacro

%macro INDIR_Y 0
   INDIR_Y_ADDR

   ; temp
   push  eax
   mem_readbyte
   pop   ebx
%endmacro

%macro INDIR_Y_BYTE 0
   INDIR_Y
%endmacro



;flags/stack macros

%macro PUSH 1   ; input in al
   mov   [ebp + STACK_OFFSET + REG_S],%1

   dec   REG_S
   and   REG_S,0x000000ff
%endmacro

%macro PULL  1
   inc   REG_S
   and   REG_S,0x000000ff

   mov   %1,[ebp + STACK_OFFSET + REG_S]
%endmacro


; Sets the Z and N flags based on given data, taken from precomputed table
%macro SET_NZ_FLAGS 1
%ifnidni %1,al
   movzx eax,%1
%else
   and   eax,0x000000ff
%endif
   and   REG_P,~(N_FLAG | Z_FLAG)
   or    REG_P,[flag_table + eax]
%endmacro


; Set/clear/test bits in the flags register
%macro SET_FLAG 1
   or    REG_P,%1
%endmacro


%macro CLEAR_FLAG 1
   and   REG_P,~(%1)
%endmacro


; For conditional jump relative instructions
; (BCC, BCS, BEQ, BMI, BNE, BPL, BVC, BVS)
%macro RELATIVE_JUMP 2
   test  REG_P,%2
   %1    %%jumptaken    ; %1 == jz/jnz
   inc   REG_PC
   ADD_CYCLES 2
   jmp   SHORT %%done
%%jumptaken:
   IMMEDIATE_BYTE
   movsx eax,al
   mov   ebx,REG_PC
   add   bl,al
   jnc   %%nopenalty
   ADD_CYCLES 1
%%nopenalty:
   ADD_CYCLES 3
   add   eax,REG_PC
   movzx REG_PC,ax
%%done:
%endmacro


%macro JUMP 1
   bank_readword %1
   mov   REG_PC,eax
%endmacro

;-----------------------------------------------------------------------
; interrupt macros
;-----------------------------------------------------------------------

%macro INT_PROC 1
   mov   ebx,REG_PC
   PUSH  bh
   PUSH  bl
   CLEAR_FLAG B_FLAG
   PUSH  REG_P
   SET_FLAG I_FLAG
   JUMP %1
%endmacro

%macro NMI_PROC 0
   INT_PROC NMI_VECTOR
%endmacro

%macro IRQ_PROC 0
   INT_PROC IRQ_VECTOR
%endmacro

%macro GET_GLOBAL_REGS 0
   ; copy 6502 regs to x86 regs
   ; a  - cl
   ; p  - ch
   ; x  - dl
   ; y  - dh
   ; s  - edi
   ; pc - esi
   mov   REG_A,[a_reg]
   mov   REG_P,[p_reg]
   mov   REG_X,[x_reg]
   mov   REG_Y,[y_reg]
   movzx dword REG_S,byte [s_reg]
   mov   REG_PC,[pc_reg]
   mov   ebp,[mem_page + 0]
%endmacro

%macro STORE_LOCAL_REGS 0
   ; copy x86 regs to 6502 regs
   ; a  - cl
   ; p  - ch
   ; x  - dl
   ; y  - dh
   ; s  - edi
   ; pc - esi
   mov   [a_reg],REG_A
   mov   [p_reg],REG_P
   mov   [x_reg],REG_X
   mov   [y_reg],REG_Y
   mov   eax,REG_S
   mov   [s_reg],al
   mov   [pc_reg],REG_PC
%endmacro


;-----------------------------------------------------------------------
; opcode macros
;-----------------------------------------------------------------------
;;
;; /*
;; ** Instruction macros
;; */
;;
%ifdef NES6502_DECIMAL
%macro ADC 2
   %2    ; read_func
   test  REG_P,D_FLAG
   jnz   %%decimal

   movzx bx,REG_P
   and   bl,C_FLAG
   add   bl,REG_A
   adc   bh,0
   add   bl,al
   jc    %%setcarry_nodec
   CLEAR_FLAG C_FLAG
   jmp   SHORT %%carrydone_nodec
%%setcarry_nodec:
   SET_FLAG C_FLAG
%%carrydone_nodec:
   xor   al,REG_A
   not   al
   mov   bh,bl    ; temp save
   xor   bl,REG_A
   and   al,bl
   test  al,0x80
   jnz   %%setoverflow_nodec
   CLEAR_FLAG V_FLAG
   jmp   SHORT %%overflowdone_nodec
%%setoverflow_nodec:
   SET_FLAG V_FLAG
%%overflowdone_nodec:
   mov   REG_A,bh
   SET_NZ_FLAGS REG_A
   jmp    %%done
%%decimal:
;;       temp = (A & 0x0F) + (data & 0x0F) + (P & C_FLAG); \
;;       if (temp >= 10) \
;;          temp = (temp - 10) | 0x10; \
;;       temp += (A & 0xF0) + (data & 0xF0); \
   mov   bl,al
   add   bl,REG_A
   and   bl,0x0F
   mov   bh,REG_P
   and   bh,C_FLAG
   add   bl,bh
   cmp   bl,9
   jle   %%nolownybblefixup
   sub   bl,10
   or    bl,0x10
%%nolownybblefixup:
   mov   ah,al
   and   ah,0xf0
   mov   bh,REG_A
   and   bh,0xf0
   add   bl,ah ; blah!
   add   bl,bh

   ; bl has our temp lovin value at this point, al is still orig data
   ; we need to preserve bl & al at this point
;;       TEST_AND_FLAG(0 == ((A + data + (P & C_FLAG)) & 0xFF), Z_FLAG); \
   mov   ah,REG_P
   and   ah,C_FLAG
   add   ah,al
   add   ah,REG_A
;   cmp   ah,0
   jz    %%setzeroflag
   CLEAR_FLAG Z_FLAG
   jmp   SHORT %%endzero
%%setzeroflag:
   SET_FLAG Z_FLAG
%%endzero:
;;       TEST_AND_FLAG(temp & 0x80, N_FLAG); \
   test  bl,0x80
   jnz   %%setnegflag
   CLEAR_FLAG N_FLAG
   jmp   SHORT %%endneg
%%setnegflag:
   SET_FLAG N_FLAG
%%endneg:
;;       TEST_AND_FLAG(((~(A ^ data)) & (A ^ temp) & 0x80), V_FLAG); \
   xor   al,REG_A
   not   al
   mov   bh,bl    ; temp save
   xor   bl,REG_A
   and   al,bl
   test  al,0x80
   jnz   %%setoverflow
   CLEAR_FLAG V_FLAG
   jmp   SHORT %%overflowdone
%%setoverflow:
   SET_FLAG V_FLAG
%%overflowdone:

;;       if (temp > 0x9F) \
;;          temp += 0x60; \
;;       TEST_AND_FLAG(temp > 0xFF, C_FLAG); \
   mov   al,bh    ; al is usually faster
   cmp   al,0x9F
   jle   %%nohinybblefixup
   add   al,0x60
   jc    %%sethinybblecarry
   CLEAR_FLAG C_FLAG
   jmp   SHORT %%nohinybblefixup
%%sethinybblecarry:
   SET_FLAG C_FLAG
%%nohinybblefixup:
;;       A = (uint8) temp; \
   mov   REG_A,al   
%%done:
   ADD_CYCLES %1
%endmacro

%else

%macro ADC 2
   %2    ; read_func
   movzx bx,REG_P
   and   bl,C_FLAG
   add   bl,REG_A
   adc   bh,0
   add   bl,al
   jc    %%setcarry
   CLEAR_FLAG C_FLAG
   jmp   SHORT %%carrydone
%%setcarry:
   SET_FLAG C_FLAG
%%carrydone:
   xor   al,REG_A
   not   al
   mov   bh,bl    ; temp save
   xor   bl,REG_A
   and   al,bl
   test  al,0x80
   jnz   %%setoverflow
   CLEAR_FLAG V_FLAG
   jmp   SHORT %%overflowdone
%%setoverflow:
   SET_FLAG V_FLAG
%%overflowdone:
   mov   REG_A,bh
   SET_NZ_FLAGS REG_A
   ADD_CYCLES %1
%endmacro
%endif
;;
;; undocumented
%macro ANC 2
   %2 ; read_func
   and   al,REG_A
   mov   REG_A,al
   SET_NZ_FLAGS al
   test  REG_P,N_FLAG
   jz    %%clear
   SET_FLAG C_FLAG
   jmp   SHORT %%done
%%clear:   
   CLEAR_FLAG C_FLAG
%%done:
   ADD_CYCLES %1
%endmacro
;;
%macro AND 2
   %2 ; read_func
   and   REG_A,al
   SET_NZ_FLAGS REG_A
   ADD_CYCLES %1
%endmacro
;;
;; undocumented
%macro ANE 2
   %2 ; read_func
   or    REG_A,0xee
   and   al,REG_X
   and   al,REG_A
   mov   REG_A,al
   SET_NZ_FLAGS al
   ADD_CYCLES %1
%endmacro
;;
;; /* undocumented */
;; #ifdef NES6502_DECIMAL
;; %macro ARR(cycles, read_func) \
;; { \
;;   read_func(data); \
;;   data &= A; \
;;   if (P & D_FLAG) \
;;   { \
;;      temp = (data >> 1) | ((P & C_FLAG) << 7); \
;;      SET_NZ_FLAGS(temp); \
;;      TEST_AND_FLAG((temp ^ data) & 0x40, V_FLAG); \
;;      if (((data & 0x0F) + (data & 0x01)) > 5) \
;;        temp = (temp & 0xF0) | ((temp + 0x6) & 0x0F); \
;;      if (((data & 0xF0) + (data & 0x10)) > 0x50) \
;;      { \
;;        temp = (temp & 0x0F) | ((temp + 0x60) & 0xF0); \
;;        SET_FLAG(C_FLAG); \
;;      } \
;;      else \
;;        CLEAR_FLAG(C_FLAG); \
;;      A = (uint8) temp; \
;;   } \
;;   else \
;;   { \
;;      A = (data >> 1) | ((P & C_FLAG) << 7); \
;;      SET_NZ_FLAGS(A); \
;;      TEST_AND_FLAG(A & 0x40, C_FLAG); \
;;      TEST_AND_FLAG(((A >> 6) ^ (A >> 5)) & 1, V_FLAG); \
;;   }\
;;   ADD_CYCLES(cycles); \
;; }
;; #else
;; %macro ARR(cycles, read_func) \
;; { \
;;   read_func(data); \
;;   data &= A; \
;;   A = (data >> 1) | ((P & C_FLAG) << 7); \
;;   SET_NZ_FLAGS(A); \
;;   TEST_AND_FLAG(A & 0x40, C_FLAG); \
;;   TEST_AND_FLAG((A >> 6) ^ (A >> 5), V_FLAG); \
;;   ADD_CYCLES(cycles); \
;; }
;; #endif /* NES6502_DECIMAL */
;;
%macro ASL 3
   %2    ; read_func
   test  al,0x80
   jnz   %%setcarry
   CLEAR_FLAG C_FLAG
   jmp   SHORT %%carrydone
%%setcarry:
   SET_FLAG C_FLAG
%%carrydone
   shl   al,1
   SET_NZ_FLAGS al
   %3    ebx,al   ; write_func
   ADD_CYCLES %1
%endmacro
;;
%macro ASL_A 0
   test  REG_A,0x80
   jz    %%clear
   SET_FLAG C_FLAG
   jmp   SHORT %%done
%%clear:
   CLEAR_FLAG C_FLAG
%%done:
   shl   REG_A,1
   SET_NZ_FLAGS REG_A
   ADD_CYCLES 2
%endmacro
;;
;; undocumented
%macro ASR 2
   %2 ; read_func
   and   REG_A,al
   test  REG_A,0x01   
   jz    %%clear
   SET_FLAG C_FLAG
   jmp   SHORT %%done
%%clear:
   CLEAR_FLAG C_FLAG
%%done:
   shr   REG_A,1
   SET_NZ_FLAGS REG_A
   ADD_CYCLES %1
%endmacro
;;
%macro BCC 0
   RELATIVE_JUMP jz,C_FLAG
%endmacro
;;
%macro BCS 0
   RELATIVE_JUMP jnz,C_FLAG
%endmacro
;;
%macro BEQ 0
   RELATIVE_JUMP jnz,Z_FLAG
%endmacro
;;
%macro BIT 2
   %2 ; read_func
   test  al,REG_A
   jz    %%setzero
   CLEAR_FLAG Z_FLAG
   jmp   SHORT %%done
%%setzero:
   SET_FLAG Z_FLAG
%%done:
   CLEAR_FLAG (N_FLAG | V_FLAG)
   and   al,(N_FLAG | V_FLAG)
   SET_FLAG al
   ADD_CYCLES %1
%endmacro
;;
%macro BMI 0
   RELATIVE_JUMP jnz,N_FLAG
%endmacro
;;
%macro BNE 0
   RELATIVE_JUMP jz,Z_FLAG
%endmacro
;;
%macro BPL 0
   RELATIVE_JUMP jz,N_FLAG
%endmacro
;;
;; /* Software interrupt type thang */
%macro BRK 0
   ; TODO: find out what PC reg hi/lo is
   mov   ebx,REG_PC
   inc   bx
   PUSH  bh
   PUSH  bl
   SET_FLAG B_FLAG
   PUSH REG_P
   SET_FLAG I_FLAG
   JUMP  IRQ_VECTOR
   and   byte [int_pending],~IRQ_MASK
   ADD_CYCLES 7
%endmacro
;;
%macro BVC 0
   RELATIVE_JUMP jz,V_FLAG
%endmacro
;;
%macro BVS 0
   RELATIVE_JUMP jnz,V_FLAG
%endmacro
;;
%macro CLC 0
   CLEAR_FLAG C_FLAG
   ADD_CYCLES 2
%endmacro
;;
%macro CLD 0
   CLEAR_FLAG D_FLAG
   ADD_CYCLES 2
%endmacro
;;
%macro CLI 0
   CLEAR_FLAG I_FLAG
   ADD_CYCLES 2
%endmacro
;;
%macro CLV 0
   CLEAR_FLAG V_FLAG
   ADD_CYCLES 2
%endmacro
;;
%macro _COMPARE 2
   sub   %2,%1
   jbe   %%setcarry
   CLEAR_FLAG C_FLAG
   jmp   SHORT %%done
%%setcarry:
   SET_FLAG C_FLAG
%%done:
   neg   al
   SET_NZ_FLAGS al
%endmacro
;;
%macro CMP 2
   %2 ; read_func
   _COMPARE REG_A,al
   ADD_CYCLES %1
%endmacro
;;
%macro CPX 2
   %2 ; read_func
   _COMPARE REG_X,al
   ADD_CYCLES %1
%endmacro
;;
%macro CPY 2
   %2 ; read_func
   _COMPARE REG_Y,al
   ADD_CYCLES %1
%endmacro
;;
;; undocumented
%macro DCP 3
   %2    ; read_func
   dec   al
   %3    ebx,al   ; write_func
   CMP   %1,EMPTY_READ
%endmacro
;;
%macro DEC 3
   %2    ; read_func
   dec   al
   %3    ebx,al
   SET_NZ_FLAGS al
   ADD_CYCLES %1
%endmacro
;;
%macro DEX 0
   dec   REG_X
   SET_NZ_FLAGS REG_X
   ADD_CYCLES 2
%endmacro
;;
%macro DEY 0
   dec   REG_Y
   SET_NZ_FLAGS REG_Y
   ADD_CYCLES 2
%endmacro 
;;
;; /* undocumented (double-NOP) */
%macro DOP 1
   inc   REG_PC
   ADD_CYCLES %1
%endmacro
;;
%macro EOR 2
   %2 ; read_func
   xor   REG_A,al
   SET_NZ_FLAGS REG_A
   ADD_CYCLES %1
%endmacro
;;
%macro INC 3
   %2    ; read_func
   inc   al
   %3    ebx,al
   SET_NZ_FLAGS al
   ADD_CYCLES %1
%endmacro
;;
%macro INX 0
   inc   REG_X
   SET_NZ_FLAGS REG_X
   ADD_CYCLES 2
%endmacro
;;
%macro INY 0
   inc   REG_Y
   SET_NZ_FLAGS REG_Y
   ADD_CYCLES 2
%endmacro
;;
;; /* undocumented */
%macro ISB 3
   %2    ; read_func
   inc   al
   %3    ebx,al   ; write_func
   SBC   %1,EMPTY_READ
%endmacro
;;
%macro JAM 0
   pushall
   cmp   dword [jam_callback],0
   je    %%nocallback
   STORE_LOCAL_REGS
   mov   eax,nes6502_context
   push  eax
   call  [eax+16*4]  ; call jam callback
   add   esp,4
   popall
   jmp   SHORT %%done
%%nocallback:
   dec   REG_PC
   mov   byte [jammed],1
   mov   byte [int_pending],0
   ADD_CYCLES 2
%%done:
%endmacro
;;
%macro JMP_INDIRECT 0
   bank_readword REG_PC
   cmp   al,0xff
   jne   %%nobug
   mov   ebx,eax
   bank_readbyte eax
   
   xchg  ebx,eax
   and   ax,~0xff
   bank_readbyte eax

   mov   bh,al
   mov   REG_PC,ebx
   jmp   SHORT %%done
%%nobug:
   JUMP  eax
%%done:
   ADD_CYCLES 5
%endmacro
;;
%macro JMP_ABSOLUTE 0
   JUMP  REG_PC
   ADD_CYCLES 3
%endmacro
;;
%macro JSR 0
   mov   ebx,REG_PC
   inc   ebx
   PUSH  bh
   PUSH  bl
   JUMP  REG_PC
   ADD_CYCLES 6
%endmacro
;;
;; undocumented
%macro LAS 2
   %2 ; read_func
   and   eax,REG_S
   mov   REG_S,eax
   mov   REG_A,al
   mov   REG_X,al
   SET_NZ_FLAGS al
   ADD_CYCLES %1
%endmacro
;;
;; undocumented
%macro LAX 2
   %2 ; read_func
   mov   REG_A,al
   mov   REG_X,al
   SET_NZ_FLAGS al
   ADD_CYCLES %1
%endmacro
;;
%macro LDA 2
   %2 ; read_func
   mov   REG_A,al
   SET_NZ_FLAGS al
   ADD_CYCLES %1
%endmacro
;;
%macro LDX 2
   %2 ; read_func
   mov   REG_X,al
   SET_NZ_FLAGS al
   ADD_CYCLES %1
%endmacro
;;
%macro LDY 2
   %2 ; read_func
   mov   REG_Y,al
   SET_NZ_FLAGS al
   ADD_CYCLES %1
%endmacro
;;
%macro LSR 3
   %2    ; read_func
   test  al,0x01
   jnz   %%setcarry
   CLEAR_FLAG C_FLAG
   jmp   SHORT %%carrydone
%%setcarry:
   SET_FLAG C_FLAG
%%carrydone:
   shr   al,1
   SET_NZ_FLAGS al
   %3    ebx,al   ; write_func
   ADD_CYCLES %1
%endmacro
;;
%macro LSR_A 0
   test  REG_A,0x01
   jnz   %%setcarry
   CLEAR_FLAG C_FLAG
   jmp   SHORT %%done
%%setcarry:
   SET_FLAG C_FLAG
%%done:
   shr   REG_A,1
   SET_NZ_FLAGS REG_A
   ADD_CYCLES 2
%endmacro
;;
;; undocumented
%macro LXA 2
   %2 ; read_func
   or    REG_A,0xee
   and   al,REG_A
   mov   REG_A,al
   mov   REG_X,al
   SET_NZ_FLAGS al
   ADD_CYCLES %1
%endmacro
;;
%macro NOP 0
   ADD_CYCLES 2
%endmacro
;;
%macro ORA 2
   %2 ; read_func
   or    REG_A,al
   SET_NZ_FLAGS REG_A
   ADD_CYCLES %1
%endmacro
;;
%macro PHA 0
   PUSH  REG_A
   ADD_CYCLES 3
%endmacro
;;
%macro PHP 0
   mov   al,REG_P
   ; B flag is pushed on stack as well
   or    al,B_FLAG
   PUSH  al
   ADD_CYCLES 3
%endmacro
;;
%macro PLA 0
   PULL  REG_A
   SET_NZ_FLAGS REG_A
   ADD_CYCLES 4
%endmacro
;;
%macro PLP 0
   PULL  REG_P
   SET_FLAG R_FLAG ; ensure reserved flag set
   ADD_CYCLES 4
%endmacro
;;
;; /* undocumented */
;; %macro RLA(cycles, read_func, write_func, addr) \
;; { \
;;   read_func(addr, data); \
;;   if (P & C_FLAG) \
;;   { \
;;      TEST_AND_FLAG(data & 0x80, C_FLAG); \
;;      data = (data << 1) | 1; \
;;   } \
;;   else \
;;   { \
;;      TEST_AND_FLAG(data & 0x80, C_FLAG); \
;;      data <<= 1; \
;;   } \
;;   write_func(addr, data); \
;;   A &= data; \
;;   SET_NZ_FLAGS(A); \
;;   ADD_CYCLES(cycles); \
;; }
;;
%macro ROL 3
   %2    ; read_func
   xor   ah,ah
   shl   ax,1
   test  REG_P,C_FLAG
   jz    %%carrynotset
   or    al,0x01
%%carrynotset:
   CLEAR_FLAG C_FLAG
   or    REG_P,ah
   SET_NZ_FLAGS al
   %3    ebx,al   ; write_func
   ADD_CYCLES %1
%endmacro
;;
%macro ROL_A 0
   mov   al,REG_A
   xor   ah,ah
   shl   ax,1
   test  REG_P,C_FLAG
   jz    %%carrynotset
   or    al,0x01
%%carrynotset:
   mov   REG_A,al
   CLEAR_FLAG C_FLAG
   or    REG_P,ah
   SET_NZ_FLAGS al
   ADD_CYCLES 2
%endmacro
;;
%macro ROR 3
   %2    ; read_func
   mov   ah,al
   xor   al,al
   shr   ax,1
   test  REG_P,C_FLAG
   jz    %%carrynotset
   or    ah,0x80
%%carrynotset:
   CLEAR_FLAG C_FLAG
   shr   al,7
   or    REG_P,al
   mov   al,ah
   SET_NZ_FLAGS al
   %3    ebx,al   ; write_func
   ADD_CYCLES %1
%endmacro
;;
%macro ROR_A 0
   mov   ah,REG_A
   xor   al,al
   shr   ax,1
   test  REG_P,C_FLAG
   jz    %%carrynotset
   or    ah,0x80
%%carrynotset:
   CLEAR_FLAG C_FLAG
   shr   al,7
   or    REG_P,al
   mov   al,ah
   mov   REG_A,al
   SET_NZ_FLAGS al
   ADD_CYCLES 2
%endmacro
;;
;; /* undocumented */
;; %macro RRA(cycles, read_func, write_func, addr) \
;; { \
;;   read_func(addr, data); \
;;   if (P & C_FLAG) \
;;   { \
;;      TEST_AND_FLAG(data & 1, C_FLAG); \
;;      data = (data >> 1) | 0x80; \
;;   } \
;;   else \
;;   { \
;;      TEST_AND_FLAG(data & 1, C_FLAG); \
;;      data >>= 1; \
;;   } \
;;   write_func(addr, data); \
;;   ADC(cycles, EMPTY_READ); \
;; }
;;
%macro RTI 0
   PULL  REG_P
   SET_FLAG R_FLAG
   xor   ebx,ebx
   PULL  bl
   PULL  bh
   mov   REG_PC,ebx
   ADD_CYCLES 6
%endmacro
;;
%macro RTS 0
   xor   ebx,ebx
   PULL  bl
   PULL  bh
   inc   bx
   mov   REG_PC,ebx
   ADD_CYCLES 6
%endmacro
;;
;; /* undocumented */
;; %macro SAX(cycles, read_func, write_func, addr) \
;; { \
;;   read_func(addr); \
;;   data = A & X; \
;;   write_func(addr, data); \
;;   ADD_CYCLES(cycles); \
;; }
;;
%ifdef NES6502_DECIMAL
;; { \
;;   read_func(data); \
;;   /* NOT(C) is considered borrow */ \
;;   temp = A - data - ((P & C_FLAG) ^ C_FLAG); \
;;   if (P & D_FLAG) \
;;   { \
;;      uint8 al, ah; \
;;      al = (A & 0x0F) - (data & 0x0F) - ((P & C_FLAG) ^ C_FLAG); \
;;      ah = (A >> 4) - (data >> 4); \
;;      if (al & 0x10) \
;;      { \
;;        al -= 6; \
;;        ah--; \
;;      } \
;;      if (ah & 0x10) \
;;        ah -= 6; \
;;      TEST_AND_FLAG(temp < 0x100, C_FLAG); \
;;      TEST_AND_FLAG(((A ^ temp) & 0x80) && ((A ^ data) & 0x80), V_FLAG); \
;;      SET_NZ_FLAGS(temp & 0xFF); \
;;      A = (ah << 4) | (al & 0x0F); \
;;   } \
;;   else \
;;   { \
;;      TEST_AND_FLAG(((A ^ temp) & 0x80) && ((A ^ data) & 0x80), V_FLAG); \
;;      TEST_AND_FLAG(temp < 0x100, C_FLAG); \
;;      A = (uint8) temp; \
;;      SET_NZ_FLAGS(A & 0xFF); \
;;   } \
;;   ADD_CYCLES(cycles); \
;; }
%macro SBC 2
%endmacro
%else
;; { \
;;   read_func(data); \
;;   temp = A - data - ((P & C_FLAG) ^ C_FLAG); \
;;   TEST_AND_FLAG(((A ^ data) & (A ^ temp) & 0x80), V_FLAG); \
;;   TEST_AND_FLAG(temp < 0x100, C_FLAG); \
;;   A = (uint8) temp; \
;;   SET_NZ_FLAGS(A); \
;;   ADD_CYCLES(cycles); \
;; }
%macro SBC 2
   %2       ; read_func
   mov   ah,REG_P
   and   ah,C_FLAG
   xor   ah,C_FLAG
   mov   bl,al    ; data save
   add   ah,al
   mov   al,REG_A
   sub   al,ah
   jnc    %%setcarry
   CLEAR_FLAG C_FLAG
   jmp   SHORT %%carrydone
%%setcarry:
   SET_FLAG C_FLAG
%%carrydone:
   mov   bh,al    ; temp save
   xor   bl,REG_A
   xor   bh,REG_A
   and   bl,bh
   test  bl,0x80
   jnz   %%setoverflow
   CLEAR_FLAG V_FLAG
   jmp   SHORT %%overflowdone
%%setoverflow:
   SET_FLAG V_FLAG
%%overflowdone:
   mov   REG_A,al
   SET_NZ_FLAGS al
   ADD_CYCLES %1
%endmacro
%endif
;;
;; /* undocumented */
;; %macro SBX(cycles, read_func) \
;; { \
;;   read_func(data); \
;;   temp = (A & X) - data; \
;;   TEST_AND_FLAG(temp < 0x100, C_FLAG); \
;;   X = temp & 0xFF; \
;;   SET_NZ_FLAGS(X); \
;;   ADD_CYCLES(cycles); \
;; }
;;
%macro SEC 0
   SET_FLAG C_FLAG
   ADD_CYCLES 2
%endmacro
;;
%macro SED 0
   SET_FLAG D_FLAG
   ADD_CYCLES 2
%endmacro
;;
%macro SEI 0
   SET_FLAG I_FLAG
   ADD_CYCLES 2
%endmacro
;;
;; /* undocumented */
;; %macro SHA(cycles, read_func, write_func, addr) \
;; { \
;;   read_func(addr); \
;;   data = A & X & ((uint8) ((addr >> 8) + 1)); \
;;   write_func(addr, data); \
;;   ADD_CYCLES(cycles); \
;; }
;;
;; /* undocumented */
;; %macro SHS(cycles, read_func, write_func, addr) \
;; { \
;;   read_func(addr); \
;;   S = A & X; \
;;   data = S & ((uint8) ((addr >> 8) + 1)); \
;;   write_func(addr, data); \
;;   ADD_CYCLES(cycles); \
;; }
;;
;; /* undocumented */
;; %macro SHX(cycles, read_func, write_func, addr) \
;; { \
;;   read_func(addr); \
;;   data = X & ((uint8) ((addr >> 8) + 1)); \
;;   write_func(addr, data); \
;;   ADD_CYCLES(cycles); \
;; }
;;
;; /* undocumented */
;; %macro SHY(cycles, read_func, write_func, addr) \
;; { \
;;   read_func(addr); \
;;   data = Y & ((uint8) ((addr >> 8 ) + 1)); \
;;   write_func(addr, data); \
;;   ADD_CYCLES(cycles); \
;; }
;;
;; /* undocumented */
;; %macro SLO(cycles, read_func, write_func, addr) \
;; { \
;;   read_func(addr, data); \
;;   TEST_AND_FLAG(data & 0x80, C_FLAG); \
;;   data <<= 1; \
;;   write_func(addr, data); \
;;   A |= data; \
;;   SET_NZ_FLAGS(A); \
;;   ADD_CYCLES(cycles); \
;; }
;;
;; /* undocumented */
;; %macro SRE(cycles, read_func, write_func, addr) \
;; { \
;;   read_func(addr, data); \
;;   TEST_AND_FLAG(data & 1, C_FLAG); \
;;   data >>= 1; \
;;   write_func(addr, data); \
;;   A ^= data; \
;;   SET_NZ_FLAGS(A); \
;;   ADD_CYCLES(cycles); \
;; }
;;
%macro STA 3
   %2 ; read_func
   %3 eax,REG_A   ; write_func addr,val
   ADD_CYCLES %1
%endmacro
;;
%macro STX 3
   %2 ; read_func
   %3 eax,REG_X   ; write_func addr,val
   ADD_CYCLES %1
%endmacro
;;
%macro STY 3
   %2 ; read_func
   %3 eax,REG_Y   ; write_func addr,val
   ADD_CYCLES %1
%endmacro
;;
%macro TAX 0
   mov   REG_X,REG_A
   SET_NZ_FLAGS REG_X
   ADD_CYCLES 2
%endmacro
;;
%macro TAY 0
   mov   REG_Y,REG_A
   SET_NZ_FLAGS REG_Y
   ADD_CYCLES 2
%endmacro
;;
;; /* undocumented (triple-NOP) */
%macro TOP 0
   add   REG_PC,2
   ADD_CYCLES 4
%endmacro
;;
%macro TSX 0
   mov   eax,REG_S
   mov   REG_X,al
   SET_NZ_FLAGS al
   ADD_CYCLES 2
%endmacro
;;
%macro TXA 0
   mov   REG_A,REG_X
   SET_NZ_FLAGS REG_A
   ADD_CYCLES 2
%endmacro
;;
%macro TXS 0
   movzx REG_S,byte REG_X
   ADD_CYCLES 2
%endmacro
;;
%macro TYA 0
   mov   REG_A,REG_Y
   SET_NZ_FLAGS REG_A
   ADD_CYCLES 2
%endmacro
;;

   SECTION  .text


   ALIGN 4
; opcode pointer table
opcode_table:
%ifdef NES6502_UNDOCUMENTED
   dd op00,op01,op02,op03,op04,op05,op06,op07,op08,op09,op0A,op0B,op0C,op0D,op0E,op0F
   dd op10,op11,op12,op13,op14,op15,op16,op17,op18,op19,op1A,op1B,op1C,op1D,op1E,op1F
   dd op20,op21,op22,op23,op24,op25,op26,op27,op28,op29,op2A,op2B,op2C,op2D,op2E,op2F
   dd op30,op31,op32,op33,op34,op35,op36,op37,op38,op39,op3A,op3B,op3C,op3D,op3E,op3F
   dd op40,op41,op42,op43,op44,op45,op46,op47,op48,op49,op4A,op4B,op4C,op4D,op4E,op4F
   dd op50,op51,op52,op53,op54,op55,op56,op57,op58,op59,op5A,op5B,op5C,op5D,op5E,op5F
   dd op60,op61,op62,op63,op64,op65,op66,op67,op68,op69,op6A,op6B,op6C,op6D,op6E,op6F
   dd op70,op71,op72,op73,op74,op75,op76,op77,op78,op79,op7A,op7B,op7C,op7D,op7E,op7F
   dd op80,op81,op82,op83,op84,op85,op86,op87,op88,op89,op8A,op8B,op8C,op8D,op8E,op8F
   dd op90,op91,op92,op93,op94,op95,op96,op97,op98,op99,op9A,op9B,op9C,op9D,op9E,op9F
   dd opA0,opA1,opA2,opA3,opA4,opA5,opA6,opA7,opA8,opA9,opAA,opAB,opAC,opAD,opAE,opAF
   dd opB0,opB1,opB2,opB3,opB4,opB5,opB6,opB7,opB8,opB9,opBA,opBB,opBC,opBD,opBE,opBF
   dd opC0,opC1,opC2,opC3,opC4,opC5,opC6,opC7,opC8,opC9,opCA,opCB,opCC,opCD,opCE,opCF
   dd opD0,opD1,opD2,opD3,opD4,opD5,opD6,opD7,opD8,opD9,opDA,opDB,opDC,opDD,opDE,opDF
   dd opE0,opE1,opE2,opE3,opE4,opE5,opE6,opE7,opE8,opE9,opEA,opEB,opEC,opED,opEE,opEF
   dd opF0,opF1,opF2,opF3,opF4,opF5,opF6,opF7,opF8,opF9,opFA,opFB,opFC,opFD,opFE,opFF
%else
   dd op00,op01,opUN,opUN,opUN,op05,op06,opUN,op08,op09,op0A,opUN,opUN,op0D,op0E,opUN
   dd op10,op11,opUN,opUN,opUN,op15,op16,opUN,op18,op19,opUN,opUN,opUN,op1D,op1E,opUN
   dd op20,op21,opUN,opUN,op24,op25,op26,opUN,op28,op29,op2A,opUN,op2C,op2D,op2E,opUN
   dd op30,op31,opUN,op33,opUN,op35,op36,opUN,op38,op39,opUN,opUN,opUN,op3D,op3E,opUN
   dd op40,op41,opUN,opUN,opUN,op45,op46,opUN,op48,op49,op4A,opUN,op4C,op4D,op4E,opUN
   dd op50,op51,opUN,opUN,opUN,op55,op56,opUN,op58,op59,opUN,opUN,opUN,op5D,op5E,opUN
   dd op60,op61,opUN,opUN,opUN,op65,op66,opUN,op68,op69,op6A,opUN,op6C,op6D,op6E,opUN
   dd op70,op71,opUN,opUN,opUN,op75,op76,opUN,op78,op79,opUN,opUN,opUN,op7D,op7E,opUN
   dd opUN,op81,opUN,opUN,op84,op85,op86,opUN,op88,opUN,op8A,opUN,op8C,op8D,op8E,opUN
   dd op90,op91,opUN,opUN,op94,op95,op96,opUN,op98,op99,op9A,opUN,opUN,op9D,opUN,opUN
   dd opA0,opA1,opA2,opUN,opA4,opA5,opA6,opUN,opA8,opA9,opAA,opUN,opAC,opAD,opAE,opUN
   dd opB0,opB1,opUN,opUN,opB4,opB5,opB6,opUN,opB8,opB9,opBA,opUN,opBC,opBD,opBE,opUN
   dd opC0,opC1,opUN,opUN,opC4,opC5,opC6,opUN,opC8,opC9,opCA,opUN,opCC,opCD,opCE,opUN
   dd opD0,opD1,opUN,opUN,opUN,opD5,opD6,opUN,opD8,opD9,opUN,opUN,opUN,opDD,opDE,opUN
   dd opE0,opE1,opUN,opUN,opE4,opE5,opE6,opUN,opE8,opE9,opEA,opUN,opEC,opED,opEE,opUN
   dd opF0,opF1,opUN,opUN,opUN,opF5,opF6,opUN,opF8,opF9,opUN,opUN,opUN,opFD,opFE,opUN
%endif

unimplemented_text:  db "unimplemented opcode $%02X\n",0
remaining_text db "remaining: %d",0

cextern exit
cextern printf

; temp macro for unimplemented opcode
%macro UNIMPLEMENTED 0
   push  eax
   mov   eax,unimplemented_text
   push  eax
   call printf
   add   esp,8
   mov   eax,-1
   push  eax
   call exit
%endmacro



%ifdef NES6502_DISASM
cextern log_printf
cextern nes6502_disasm

%macro DISASSEMBLE 0
   push  ecx
   push  edx
   push  esi
   push  edi

   mov   eax,REG_S
   and   eax,0x000000ff
   push  eax
   mov   al,REG_Y
   push  eax
   mov   al,REG_X
   push  eax
   mov   al,REG_A
   push  eax
   mov   al,REG_P
   push  eax
   push  REG_PC
   call  _nes6502_disasm
   add   esp,24   ; junk off stack

   push  eax
   call  _log_printf
   add   esp,4 ; drop junk off stack
   
   pop   edi
   pop   esi
   pop   edx
   pop   ecx
%endmacro
%endif

   ALIGN 4
;-----------------------------------------------------------------------
; opcode functions
;-----------------------------------------------------------------------
%macro OPCODE_END 0
%ifdef NES6502_DISASM
   DISASSEMBLE
%endif
   cmp   dword [remaining_cycles], 0
   jle   NEAR end_execute
   bank_readbyte REG_PC
   inc   REG_PC
;   and   eax,0x000000ff
   jmp   [opcode_table + eax*4]
%endmacro

op00:  ; BRK
   BRK
OPCODE_END


op01:  ; ORA ($nn,X)
   ORA   6,INDIR_X_BYTE
OPCODE_END

op02:  ; JAM
op12:  ; JAM
op22:  ; JAM
op32:  ; JAM
op42:  ; JAM
op52:  ; JAM
op62:  ; JAM
op72:  ; JAM
op92:  ; JAM
opB2:  ; JAM
opD2:  ; JAM
opF2:  ; JAM
   JAM
   xor   eax,eax
   mov   [remaining_cycles],eax
OPCODE_END

op03:  ; SLO ($nn,X)
;   SLO(8, INDIR_X, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END

op04:  ; NOP $nn
op44:  ; NOP $nn
op64:  ; NOP $nn
   DOP 3
OPCODE_END

op05:  ; ORA $nn
   ORA   3,ZERO_PAGE_BYTE
OPCODE_END

op06:  ; ASL $nn
   ASL   5,ZERO_PAGE,ZP_WRITEBYTE
OPCODE_END


op07:  ; SLO $nn
;   SLO(5, ZERO_PAGE, ZP_WRITEBYTE, baddr)
   UNIMPLEMENTED
OPCODE_END


op08:  ; PHP
   PHP
OPCODE_END


op09:  ; ORA #$nn
   ORA   2,IMMEDIATE_BYTE
OPCODE_END


op0A:  ; ASL A
   ASL_A
OPCODE_END


op0B:  ; ANC #$nn
   ANC   2,IMMEDIATE_BYTE
OPCODE_END


op0C:  ; NOP $nnnn
   TOP
OPCODE_END


op0D:  ; ORA $nnnn
   ORA   4,ABSOLUTE_BYTE
OPCODE_END


op0E:  ; ASL $nnnn
   ASL   6,ABSOLUTE,MEMORY_WRITEBYTE
OPCODE_END


op0F:  ; SLO $nnnn
;   SLO(6, ABSOLUTE, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op10:  ; BPL $nnnn
   BPL
OPCODE_END


op11:  ; ORA ($nn),Y
   ORA   5,INDIR_Y_BYTE
OPCODE_END


op13:  ; SLO ($nn),Y
;   SLO(8, INDIR_Y, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op14:  ; NOP $nn,X
op34:  ; NOP
op54:  ; NOP $nn,X
op74:  ; NOP $nn,X
opD4:  ; NOP $nn,X
opF4:  ; NOP ($nn,X)
   DOP 4
OPCODE_END


op15:  ; ORA $nn,X
   ORA   4,ZP_IND_X_BYTE
OPCODE_END


op16:  ; ASL $nn,X
   ASL   6,ZP_IND_X,ZP_WRITEBYTE
OPCODE_END


op17:  ; SLO $nn,X
;   SLO(6, ZP_IND_X, ZP_WRITEBYTE, baddr)
   UNIMPLEMENTED
OPCODE_END


op18:  ; CLC
   CLC
OPCODE_END


op19:  ; ORA $nnnn,Y
   ORA   4,ABS_IND_Y_BYTE
OPCODE_END


op1A:  ; NOP
op3A:  ; NOP
op5A:  ; NOP
op7A:  ; NOP
opDA:  ; NOP
opFA:  ; NOP
   NOP
OPCODE_END


op1B:  ; SLO $nnnn,Y
;   SLO(7, ABS_IND_Y, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op1C:  ; NOP $nnnn,X
op3C:  ; NOP $nnnn,X
op5C:  ; NOP $nnnn,X
op7C:  ; NOP $nnnn,X
opDC:  ; NOP $nnnn,X
opFC:  ; NOP $nnnn,X
   TOP
OPCODE_END


op1D:  ; ORA $nnnn,X
   ORA   4,ABS_IND_X_BYTE
OPCODE_END


op1E:  ; ASL $nnnn,X
   ASL   7,ABS_IND_X,MEMORY_WRITEBYTE
OPCODE_END


op1F:  ; SLO $nnnn,X
;   SLO(7, ABS_IND_X, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op20:  ; JSR $nnnn
   JSR
OPCODE_END


op21:  ; AND ($nn,X)
   AND   6,INDIR_X_BYTE
OPCODE_END


op23:  ; RLA ($nn,X)
;   RLA(8, INDIR_X, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op24:  ; BIT $nn
   BIT   3,ZERO_PAGE_BYTE
OPCODE_END


op25:  ; AND $nn
   AND   3,ZERO_PAGE_BYTE
OPCODE_END


op26:  ; ROL $nn
   ROL   5,ZERO_PAGE,ZP_WRITEBYTE
OPCODE_END


op27:  ; RLA $nn
;   RLA(5, ZERO_PAGE, ZP_WRITEBYTE, baddr)
   UNIMPLEMENTED
OPCODE_END


op28:  ; PLP
   PLP
OPCODE_END


op29:  ; AND #$nn
   AND   2,IMMEDIATE_BYTE
OPCODE_END


op2A:  ; ROL A
   ROL_A
OPCODE_END


op2B:  ; ANC #$nn
   ANC   2,IMMEDIATE_BYTE
OPCODE_END


op2C:  ; BIT $nnnn
   BIT   4,ABSOLUTE_BYTE
OPCODE_END


op2D:  ; AND $nnnn
   AND   4,ABSOLUTE_BYTE
OPCODE_END


op2E:  ; ROL $nnnn
   ROL   6,ABSOLUTE,MEMORY_WRITEBYTE
OPCODE_END


op2F:  ; RLA $nnnn
;   RLA(6, ABSOLUTE, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op30:  ; BMI $nnnn
   BMI
OPCODE_END


op31:  ; AND ($nn),Y
   AND   5,INDIR_Y_BYTE
OPCODE_END


op33:  ; RLA ($nn),Y
;   RLA(8, INDIR_Y, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op35:  ; AND $nn,X
   AND   4,ZP_IND_X_BYTE
OPCODE_END


op36:  ; ROL $nn,X
   ROL   6,ZP_IND_X,ZP_WRITEBYTE
OPCODE_END


op37:  ; RLA $nn,X
;   RLA(6, ZP_IND_X, ZP_WRITEBYTE, baddr)
   UNIMPLEMENTED
OPCODE_END


op38:  ; SEC
   SEC
OPCODE_END


op39:  ; AND $nnnn,Y
   AND   4,ABS_IND_Y_BYTE
OPCODE_END


op3B:  ; RLA $nnnn,Y
;   RLA(7, ABS_IND_Y, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op3D:  ; AND $nnnn,X
   AND   4,ABS_IND_X_BYTE
OPCODE_END


op3E:  ; ROL $nnnn,X
   ROL   7,ABS_IND_X,MEMORY_WRITEBYTE
OPCODE_END


op3F:  ; RLA $nnnn,X
;   RLA(7, ABS_IND_X, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op40:  ; RTI
   RTI
OPCODE_END


op41:  ; EOR ($nn,X)
   EOR   6,INDIR_X_BYTE
OPCODE_END


op43:  ; SRE ($nn,X)
;   SRE(8, INDIR_X, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op45:  ; EOR $nn
   EOR   3,ZERO_PAGE_BYTE
OPCODE_END


op46:  ; LSR $nn
   LSR   5,ZERO_PAGE,ZP_WRITEBYTE
OPCODE_END


op47:  ; SRE $nn
;   SRE(5, ZERO_PAGE, ZP_WRITEBYTE, baddr)
   UNIMPLEMENTED
OPCODE_END


op48:  ; PHA
   PHA
OPCODE_END


op49:  ; EOR #$nn
   EOR   2,IMMEDIATE_BYTE
OPCODE_END


op4A:  ; LSR A
   LSR_A
OPCODE_END


op4B:  ; ASR #$nn
   ASR   2,IMMEDIATE_BYTE
OPCODE_END


op4C:  ; JMP $nnnn
   JMP_ABSOLUTE
OPCODE_END


op4D:  ; EOR $nnnn
   EOR   4,ABSOLUTE_BYTE
OPCODE_END


op4E:  ; LSR $nnnn
   LSR   6,ABSOLUTE,MEMORY_WRITEBYTE
OPCODE_END


op4F:  ; SRE $nnnn
;   SRE(6, ABSOLUTE, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op50:  ; BVC $nnnn
   BVC
OPCODE_END


op51:  ; EOR ($nn),Y
   EOR   5,INDIR_Y_BYTE
OPCODE_END


op53:  ; SRE ($nn),Y
;   SRE(8, INDIR_Y, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op55:  ; EOR $nn,X
   EOR   4,ZP_IND_X_BYTE
OPCODE_END


op56:  ; LSR $nn,X
   LSR   6,ZP_IND_X,ZP_WRITEBYTE
OPCODE_END


op57:  ; SRE $nn,X
;   SRE(6, ZP_IND_X, ZP_WRITEBYTE, baddr)
   UNIMPLEMENTED
OPCODE_END


op58:  ; CLI
   CLI
OPCODE_END


op59:  ; EOR $nnnn,Y
   EOR   4,ABS_IND_Y_BYTE
OPCODE_END


op5B:  ; SRE $nnnn,Y
;   SRE(7, ABS_IND_Y, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op5D:  ; EOR $nnnn,X
   EOR   4,ABS_IND_X_BYTE
OPCODE_END


op5E:  ; LSR $nnnn,X
   LSR   7,ABS_IND_X,MEMORY_WRITEBYTE
OPCODE_END


op5F:  ; SRE $nnnn,X
;   SRE(7, ABS_IND_X, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op60:  ; RTS
   RTS
OPCODE_END


op61:  ; ADC ($nn,X)
   ADC   6,INDIR_X_BYTE
OPCODE_END


op63:  ; RRA ($nn,X)
;   RRA(8, INDIR_X, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op65:  ; ADC $nn
   ADC   3,ZERO_PAGE_BYTE
OPCODE_END


op66:  ; ROR $nn
   ROR   5,ZERO_PAGE,ZP_WRITEBYTE
OPCODE_END


op67:  ; RRA $nn
;   RRA(5, ZERO_PAGE, ZP_WRITEBYTE, baddr)
   UNIMPLEMENTED
OPCODE_END


op68:  ; PLA
   PLA
OPCODE_END


op69:  ; ADC #$nn
   ADC   2,IMMEDIATE_BYTE
OPCODE_END


op6A:  ; ROR A
   ROR_A
OPCODE_END


op6B:  ; ARR #$nn
;   ARR(2, IMMEDIATE_BYTE)
   UNIMPLEMENTED
OPCODE_END


op6C:  ; JMP ($nnnn)
   JMP_INDIRECT
OPCODE_END


op6D:  ; ADC $nnnn
   ADC   4,ABSOLUTE_BYTE
OPCODE_END


op6E:  ; ROR $nnnn
   ROR   6,ABSOLUTE,MEMORY_WRITEBYTE
OPCODE_END


op6F:  ; RRA $nnnn
;   RRA(6, ABSOLUTE, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op70:  ; BVS $nnnn
   BVS
OPCODE_END


op71:  ; ADC ($nn),Y
   ADC   5,INDIR_Y_BYTE
OPCODE_END


op73:  ; RRA ($nn),Y
;   RRA(8, INDIR_Y, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op75:  ; ADC $nn,X
   ADC   4,ZP_IND_X_BYTE
OPCODE_END


op76:  ; ROR $nn,X
   ROR   6,ZP_IND_X,ZP_WRITEBYTE
OPCODE_END


op77:  ; RRA $nn,X
;   RRA(6, ZP_IND_X, ZP_WRITEBYTE, baddr)
   UNIMPLEMENTED
OPCODE_END


op78:  ; SEI
   SEI
OPCODE_END


op79:  ; ADC $nnnn,Y
   ADC   4,ABS_IND_Y_BYTE
OPCODE_END


op7B:  ; RRA $nnnn,Y
;   RRA(7, ABS_IND_Y, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op7D:  ; ADC $nnnn,X
   ADC   4,ABS_IND_X_BYTE
OPCODE_END


op7E:  ; ROR $nnnn,X
   ROR   7,ABS_IND_X,MEMORY_WRITEBYTE
OPCODE_END


op7F:  ; RRA $nnnn,X
;   RRA(7, ABS_IND_X, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op80:  ; NOP #$nn
op82:  ; NOP #$nn
op89:  ; NOP #$nn
opC2:  ; NOP #$nn
opE2:  ; NOP #$nn
   DOP 2
OPCODE_END


op81:  ; STA ($nn,X)
   STA   6,INDIR_X_ADDR,MEMORY_WRITEBYTE
OPCODE_END


op83:  ; SAX ($nn,X)
;   SAX(6, INDIR_X_ADDR, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op84:  ; STY $nn
   STY   3,ZERO_PAGE_ADDR,ZP_WRITEBYTE
OPCODE_END


op85:  ; STA $nn
   STA   3,ZERO_PAGE_ADDR,ZP_WRITEBYTE
OPCODE_END


op86:  ; STX $nn
   STX   3,ZERO_PAGE_ADDR,ZP_WRITEBYTE
OPCODE_END


op87:  ; SAX $nn
;   SAX(3, ZERO_PAGE_ADDR, ZP_WRITEBYTE, baddr)
   UNIMPLEMENTED
OPCODE_END


op88:  ; DEY
   DEY
OPCODE_END


op8A:  ; TXA
   TXA
OPCODE_END


op8B:  ; ANE #$nn
   ANE   2,IMMEDIATE_BYTE
OPCODE_END


op8C:  ; STY $nnnn
   STY   4,ABSOLUTE_ADDR,MEMORY_WRITEBYTE
OPCODE_END


op8D:  ; STA $nnnn
   STA   4,ABSOLUTE_ADDR,MEMORY_WRITEBYTE
OPCODE_END


op8E:  ; STX $nnnn
   STX   4,ABSOLUTE_ADDR,MEMORY_WRITEBYTE
OPCODE_END


op8F:  ; SAX $nnnn
;   SAX(4, ABSOLUTE_ADDR, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op90:  ; BCC $nnnn
   BCC
OPCODE_END


op91:  ; STA ($nn),Y
   STA   6,INDIR_Y_ADDR,MEMORY_WRITEBYTE
OPCODE_END


op93:  ; SHA ($nn),Y
;   SHA(6, INDIR_Y_ADDR, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op94:  ; STY $nn,X
   STY   4,ZP_IND_X_ADDR,ZP_WRITEBYTE
OPCODE_END


op95:  ; STA $nn,X
   STA   4,ZP_IND_X_ADDR,ZP_WRITEBYTE
OPCODE_END


op96:  ; STX $nn,Y
   STX   4,ZP_IND_Y_ADDR,ZP_WRITEBYTE
OPCODE_END


op97:  ; SAX $nn,Y
;   SAX(4, ZP_IND_Y_ADDR, ZP_WRITEBYTE, baddr)
   UNIMPLEMENTED
OPCODE_END


op98:  ; TYA
   TYA
OPCODE_END


op99:  ; STA $nnnn,Y
   STA   5,ABS_IND_Y_ADDR,MEMORY_WRITEBYTE
OPCODE_END


op9A:  ; TXS
   TXS
OPCODE_END


op9B:  ; SHS $nnnn,Y
;   SHS(5, ABS_IND_Y_ADDR, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op9C:  ; SHY $nnnn,X
;   SHY(5, ABS_IND_X_ADDR, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op9D:  ; STA $nnnn,X
   STA   5,ABS_IND_X_ADDR,MEMORY_WRITEBYTE
OPCODE_END


op9E:  ; SHX $nnnn,Y
;   SHX(5, ABS_IND_Y_ADDR, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


op9F:  ; SHA $nnnn,Y
;   SHA(5, ABS_IND_Y_ADDR, MEMORY_WRITEBYTE, addr)
   UNIMPLEMENTED
OPCODE_END


opA0:  ; LDY #$nn
   LDY   2,IMMEDIATE_BYTE
OPCODE_END


opA1:  ; LDA ($nn,X)
   LDA   6,INDIR_X_BYTE
OPCODE_END


opA2:  ; LDX #$nn
   LDX   2,IMMEDIATE_BYTE
OPCODE_END


opA3:  ; LAX ($nn,X)
   LAX   6,INDIR_X_BYTE
OPCODE_END


opA4:  ; LDY $nn
   LDY   3,ZERO_PAGE_BYTE
OPCODE_END


opA5:  ; LDA $nn
   LDA   3,ZERO_PAGE_BYTE
OPCODE_END


opA6:  ; LDX $nn
   LDX   3,ZERO_PAGE_BYTE
OPCODE_END


opA7:  ; LAX $nn
   LAX   3,ZERO_PAGE_BYTE
OPCODE_END


opA8:  ; TAY
   TAY
OPCODE_END


opA9:  ; LDA #$nn
   LDA   2,IMMEDIATE_BYTE
OPCODE_END


opAA:  ; TAX
   TAX
OPCODE_END


opAB:  ; LXA #$nn
   LXA   2,IMMEDIATE_BYTE
OPCODE_END


opAC:  ; LDY $nnnn
   LDY   4,ABSOLUTE_BYTE
OPCODE_END


opAD:  ; LDA $nnnn
   LDA   4,ABSOLUTE_BYTE
OPCODE_END


opAE:  ; LDX $nnnn
   LDX   4,ABSOLUTE_BYTE
OPCODE_END


opAF:  ; LAX $nnnn
   LAX   4,ABSOLUTE_BYTE
OPCODE_END


opB0:  ; BCS $nnnn
   BCS
OPCODE_END


opB1:  ; LDA ($nn),Y
   LDA   5,INDIR_Y_BYTE
OPCODE_END


opB3:  ; LAX ($nn),Y
   LAX   5,INDIR_Y_BYTE
OPCODE_END


opB4:  ; LDY $nn,X
   LDY   4,ZP_IND_X_BYTE
OPCODE_END


opB5:  ; LDA $nn,X
   LDA   4,ZP_IND_X_BYTE
OPCODE_END


opB6:  ; LDX $nn,Y
   LDX   4,ZP_IND_Y_BYTE
OPCODE_END


opB7:  ; LAX $nn,Y
   LAX   4,ZP_IND_Y_BYTE
OPCODE_END


opB8:  ; CLV
   CLV
OPCODE_END


opB9:  ; LDA $nnnn,Y
   LDA   4,ABS_IND_Y_BYTE
OPCODE_END


opBA:  ; TSX
   TSX
OPCODE_END


opBB:  ; LAS $nnnn,Y
   LAS   4,ABS_IND_Y_BYTE
OPCODE_END


opBC:  ; LDY $nnnn,X
   LDY   4,ABS_IND_X_BYTE
OPCODE_END


opBD:  ; LDA $nnnn,X
   LDA   4,ABS_IND_X_BYTE
OPCODE_END


opBE:  ; LDX $nnnn,Y
   LDX   4,ABS_IND_Y_BYTE
OPCODE_END


opBF:  ; LAX $nnnn,Y
   LAX   4,ABS_IND_Y_BYTE
OPCODE_END


opC0:  ; CPY #$nn
   CPY   2,IMMEDIATE_BYTE
OPCODE_END


opC1:  ; CMP ($nn,X)
   CMP   6,INDIR_X_BYTE
OPCODE_END


opC3:  ; DCP ($nn,X)
   DCP   8,INDIR_X,MEMORY_WRITEBYTE
OPCODE_END


opC4:  ; CPY $nn
   CPY   3,ZERO_PAGE_BYTE
OPCODE_END


opC5:  ; CMP $nn
   CMP   3,ZERO_PAGE_BYTE
OPCODE_END


opC6:  ; DEC $nn
   DEC   5,ZERO_PAGE,ZP_WRITEBYTE
OPCODE_END


opC7:  ; DCP $nn
   DCP   5,ZERO_PAGE,ZP_WRITEBYTE
OPCODE_END


opC8:  ; INY
   INY
OPCODE_END


opC9:  ; CMP #$nn
   CMP   2,IMMEDIATE_BYTE
OPCODE_END


opCA:  ; DEX
   DEX
OPCODE_END


opCB:  ; SBX #$nn
;   SBX(2, IMMEDIATE_BYTE)
   UNIMPLEMENTED
OPCODE_END


opCC:  ; CPY $nnnn
   CPY   4,ABSOLUTE_BYTE
OPCODE_END


opCD:  ; CMP $nnnn
   CMP   4,ABSOLUTE_BYTE
OPCODE_END


opCE:  ; DEC $nnnn
   DEC   6,ABSOLUTE,MEMORY_WRITEBYTE
OPCODE_END


opCF:  ; DCP $nnnn
   DCP   6,ABSOLUTE,MEMORY_WRITEBYTE
OPCODE_END


opD0:  ; BNE $nnnn
   BNE
OPCODE_END


opD1:  ; CMP ($nn),Y
   CMP   5,INDIR_Y_BYTE
OPCODE_END


opD3:  ; DCP ($nn),Y
   DCP   8,INDIR_Y,MEMORY_WRITEBYTE
OPCODE_END


opD5:  ; CMP $nn,X
   CMP   4,ZP_IND_X_BYTE
OPCODE_END


opD6:  ; DEC $nn,X
   DEC   6,ZP_IND_X,ZP_WRITEBYTE
OPCODE_END


opD7:  ; DCP $nn,X
   DCP   6,ZP_IND_X,ZP_WRITEBYTE
OPCODE_END


opD8:  ; CLD
   CLD
OPCODE_END


opD9:  ; CMP $nnnn,Y
   CMP   4,ABS_IND_Y_BYTE
OPCODE_END


opDB:  ; DCP $nnnn,Y
   DCP   7,ABS_IND_Y,MEMORY_WRITEBYTE
OPCODE_END


opDD:  ; CMP $nnnn,X
   CMP   4,ABS_IND_X_BYTE
OPCODE_END


opDE:  ; DEC $nnnn,X
   DEC   7,ABS_IND_X,MEMORY_WRITEBYTE
OPCODE_END


opDF:  ; DCP $nnnn,X
   DCP   7,ABS_IND_X,MEMORY_WRITEBYTE
OPCODE_END


opE0:  ; CPX #$nn
   CPX   2,IMMEDIATE_BYTE
OPCODE_END


opE1:  ; SBC ($nn,X)
   SBC   6,INDIR_X_BYTE
OPCODE_END


opE3:  ; ISB ($nn,X)
   ISB   8,INDIR_X,MEMORY_WRITEBYTE
OPCODE_END


opE4:  ; CPX $nn
   CPX   3,ZERO_PAGE_BYTE
OPCODE_END


opE5:  ; SBC $nn
   SBC   3,ZERO_PAGE_BYTE
OPCODE_END


opE6:  ; INC $nn
   INC   5,ZERO_PAGE,ZP_WRITEBYTE
OPCODE_END


opE7:  ; ISB $nn
   ISB   5,ZERO_PAGE,ZP_WRITEBYTE
OPCODE_END


opE8:  ; INX
   INX
OPCODE_END


opE9:  ; SBC #$nn
opEB:  ; USBC #$nn
   SBC   2,IMMEDIATE_BYTE
OPCODE_END


opEA:  ; NOP
   NOP
OPCODE_END


opEC:  ; CPX $nnnn
   CPX   4,ABSOLUTE_BYTE
OPCODE_END


opED:  ; SBC $nnnn
   SBC   4,ABSOLUTE_BYTE
OPCODE_END


opEE:  ; INC $nnnn
   INC   6,ABSOLUTE,MEMORY_WRITEBYTE
OPCODE_END


opEF:  ; ISB $nnnn
   ISB   6,ABSOLUTE,MEMORY_WRITEBYTE
OPCODE_END


opF0:  ; BEQ $nnnn
   BEQ
OPCODE_END


opF1:  ; SBC ($nn),Y
   SBC   5,INDIR_Y_BYTE
OPCODE_END


opF3:  ; ISB ($nn),Y
   ISB   8,INDIR_Y,MEMORY_WRITEBYTE
OPCODE_END


opF5:  ; SBC $nn,X
   SBC   4,ZP_IND_X_BYTE
OPCODE_END


opF6:  ; INC $nn,X
   INC   6,ZP_IND_X,ZP_WRITEBYTE
OPCODE_END


opF7:  ; ISB $nn,X
   ISB   6,ZP_IND_X,ZP_WRITEBYTE
OPCODE_END


opF8:  ; SED
   SED
OPCODE_END


opF9:  ; SBC $nnnn,Y
   SBC   4,ABS_IND_Y_BYTE
OPCODE_END


opFB:  ; ISB $nnnn,Y
   ISB   7,ABS_IND_Y,MEMORY_WRITEBYTE
OPCODE_END


opFD:  ; SBC $nnnn,X
   SBC   4,ABS_IND_X_BYTE
OPCODE_END


opFE:  ; INC $nnnn,X
   INC   7,ABS_IND_X,MEMORY_WRITEBYTE
OPCODE_END


opFF:  ; ISB $nnnn,X
   ISB   7,ABS_IND_X,MEMORY_WRITEBYTE
OPCODE_END


%ifndef NES6502_UNDOCUMENTED
opUN: ; undocumented opcodes not emulated
   NOP
OPCODE_END
%endif


   ALIGN 4
;-----------------------------------------------------------------------
; support functions
;-----------------------------------------------------------------------
%if 0
cglobal nes6502_init
nes6502_init:
   mov   byte [a_reg],0
   mov   byte [x_reg],0
   mov   byte [y_reg],0
   mov   byte [s_reg],0xff
   ret
%endif

   ALIGN 4
;-----------------------------------------------------------------------
cglobal nes6502_reset
nes6502_reset:
   push  ebx
   mov   byte [p_reg],Z_FLAG | R_FLAG | I_FLAG
   xor   eax,eax
   mov   [int_pending],al
   mov   [jammed],al
   mov   [total_cycles],eax
   bank_readword RESET_VECTOR
   mov   [pc_reg],eax
   mov   byte [burn_cycles],RESET_CYCLES
   pop   ebx
   ret

   ALIGN 4
;-----------------------------------------------------------------------
cglobal nes6502_nmi
nes6502_nmi:
   cmp   byte [jammed],0
   jz    .do_nmi
   ret

.do_nmi:
   pushall
   GET_GLOBAL_REGS
   NMI_PROC
   add   [burn_cycles], byte INT_CYCLES
   STORE_LOCAL_REGS
   popall
   ret

   ALIGN 4
;-----------------------------------------------------------------------
cglobal nes6502_irq
nes6502_irq:
   cmp   byte [jammed],0
   jz    .do_irq
   ret

.do_irq:
   pushall
   GET_GLOBAL_REGS

   test  REG_P,I_FLAG
   jnz   .no_dice

   IRQ_PROC
   add   [burn_cycles], byte INT_CYCLES
   jmp   SHORT .irq_done

.no_dice:
   mov   [int_pending], byte 1

.irq_done:
   STORE_LOCAL_REGS
   popall
   ret

   ALIGN 4
;-----------------------------------------------------------------------
cglobal nes6502_burn
nes6502_burn:
   mov   eax,[esp+4]    ; get param1 = dma_cycles to add
   add   [burn_cycles],eax
   ret

   ALIGN 4
;-----------------------------------------------------------------------
;set the current context
cglobal nes6502_setcontext
nes6502_setcontext:
   mov   eax,[esp+4]    ; get param1 = pointer to context
   push  edi
   push  esi
   push  ecx
   mov   ecx,nes6502_context_end - nes6502_context
   mov   esi,eax
   mov   edi,nes6502_context
   rep   movsb
   pop   ecx
   pop   esi
   pop   edi
   mov   byte [jammed],0
   ret

   ALIGN 4
;-----------------------------------------------------------------------
;get the current context
cglobal nes6502_getcontext
nes6502_getcontext:
   mov   eax,[esp+4]    ; get param1 = pointer to context
   push  edi
   push  esi
   push  ecx
   mov   ecx,nes6502_context_end - nes6502_context
   mov   edi,eax
   mov   esi,nes6502_context
   rep   movsb
   pop   ecx
   pop   esi
   pop   edi
   ret

   ALIGN 4
;-----------------------------------------------------------------------
cglobal nes6502_getbyte
nes6502_getbyte:
   mov   eax,[esp + 4]  ; param1 = address

   ; we clobber ebx
   push  ebx
   bank_readbyte eax
   pop   ebx
   ret

   ALIGN 4
;-----------------------------------------------------------------------
cglobal nes6502_getcycles
nes6502_getcycles:
   mov   eax,[esp + 4]  ; get reset parameter
   or    eax,eax
   jz    .no_reset
   xor   eax,eax
   xchg  eax,[total_cycles]
   ret

.no_reset:
   mov   eax,[total_cycles]
   ret


   ALIGN 4
;-----------------------------------------------------------------------
cglobal nes6502_release
nes6502_release:
   mov   [remaining_cycles], dword 0
   ret


   ALIGN 4
;-----------------------------------------------------------------------
cglobal nes6502_clearirq
nes6502_clearirq:
   mov   [int_pending], byte 0
   ret


;-----------------------------------------------------------------------

   ALIGN 4
;-----------------------------------------------------------------------
;-----------------------------------------------------------------------
;main
;-----------------------------------------------------------------------
;-----------------------------------------------------------------------
cglobal nes6502_execute
nes6502_execute:
   mov   eax,[esp + 4]     ; get remaining cycles
   mov   [remaining_cycles],eax

   mov   eax,[total_cycles]
   push  eax   ; save initial cycle count

   pushall
   GET_GLOBAL_REGS

   ; zero ebx,eax
   xor   ebx,ebx
   xor   eax,eax

%if 0

   cmp   dword [remaining_cycles],eax
   jle   .dont_do_irq

   mov   ah,[int_pending]
   or    ah,ah
   jz    .dont_do_irq

   test  REG_P,I_FLAG
   jnz   .dont_do_irq

   ; al == 0 here
   mov   [int_pending],al

   IRQ_PROC
   ADD_CYCLES INT_CYCLES

.dont_do_irq:
   cmp   dword [remaining_cycles], 0
   jle   .dont_do_dma

   mov   eax,[burn_cycles]
   or    eax,eax
   jz    .dont_do_dma

   ; burn_cycles>0
   mov   ebx,eax
   cmp   [remaining_cycles],eax
   ja    .remain_bigger
   mov   ebx,[remaining_cycles]
.remain_bigger:
   sub   [remaining_cycles],ebx
   add   [total_cycles],ebx
   sub   [burn_cycles],ebx

.dont_do_dma:

%endif

   cmp   dword [remaining_cycles],eax
   jle   .dont_do_dma

   mov   eax,[burn_cycles]
   or    eax,eax
   jz    .dont_do_dma

   ; burn_cycles>0
   mov   ebx,eax
   cmp   [remaining_cycles],eax
   ja    .remain_bigger
   mov   ebx,[remaining_cycles]
.remain_bigger:
   sub   [remaining_cycles],ebx
   add   [total_cycles],ebx
   sub   [burn_cycles],ebx

.dont_do_dma:
   xor   eax,eax
   cmp   dword [remaining_cycles],eax
   jle   .dont_do_irq

   cmp   byte [int_pending],al
   jz    .dont_do_irq

   test  REG_P,I_FLAG
   jnz   .dont_do_irq

   ; al == 0 here
   mov   [int_pending],al

   IRQ_PROC
   ADD_CYCLES INT_CYCLES

.dont_do_irq:

   OPCODE_END

end_execute:
   
   STORE_LOCAL_REGS
   popall

   ; Return our actual amount of executed cycles
   ; return (total_cycles - old_cycles);
   pop   eax
   sub   eax,[total_cycles]
   ; negate...
   neg   eax
   ret


; $Log: $
