//
// shatbox (C) 2002 Matthew Conte (matt@conte.com)
//
//
// ppu.cpp
//
// ppu emulation
//
// $Id: $
//

#include <stdio.h>
#include <string.h>
#include "ppu.h"
#include "cpu.h"
#include "input.h"
#include "oam.h"
#include "nes.h"


#define  BYTES_PER_TILE       16
#define  PPU_ACCESS(x)        (mpPage[(x) / PPU_BANKSIZE][(x)])


static const nes6502_memwrite ppu_memwrite[] =
{
   { 0x2000, 0x3fff, ppu::dispatchWrite, NULL },
   { 0x4014, 0x4014, ppu::dispatchWriteDMA, NULL },
   { (uint32) -1, (uint32) -1, NULL, NULL }
};

static const nes6502_memread ppu_memread[] =
{
   { 0x2000, 0x3fff, ppu::dispatchRead, NULL },
   { (uint32) -1, (uint32) -1, NULL, NULL }
};

ppu::ppu(machine *pMachine)
: unit(pMachine)
{
   mpMemWrite = ppu_memwrite;
   mpMemRead = ppu_memread;

#ifndef NO_TILECACHE
   memset(mTileCache, 0, sizeof(mTileCache));
   memset(mTileDirty, false, sizeof(mTileDirty));
   mAnyTilesDirty = false;
   memset(mPageDirty, false, sizeof(mPageDirty));
   mAnyPagesDirty = false;
#endif

   // write protect all low pages
   for (int i = 0; i < 16; i++)
   {
      mpPage[i] = NULL;
      mProtect[i] = (i < 8) ? true : false;
   }

   memset(mNameTable, 0, sizeof(mNameTable));

   mPalette.setHue(346.0f);
   mPalette.setTint(0.43f);
   mPalette.generate(mColors);

   memset(mRegs, 0, sizeof(mRegs));

   mVAddr = 0;
   mVAddrLatch = 0;
   mTileXOfs = 0;
   mFlipFlop = false;

   mScanline = 0;
   mLatch = 0;
   mVDataLatch = 0;

   mStrikeCycle = 0;
   mStrike = false;

   mpOAM = new oam;
}


ppu::~ppu()
{
   delete mpOAM;
}


void ppu::reset(bool hardReset)
{
   mpOAM->reset(hardReset);

#ifndef NO_TILECACHE
   // dirty all pages
   mAnyPagesDirty = true;
   memset(mPageDirty, true, sizeof(mPageDirty));

   mAnyTilesDirty = false;
   memset(mTileDirty, false, sizeof(mTileDirty));
#endif // !NO_TILECACHE

   mFlipFlop = false;
   mVAddr = 0x2000;
   mVAddrLatch = 0x2000;
   mLatch = 0;

   // start us in vblank.
   mScanline = 241;

   memset(mRegs, 0, sizeof(mRegs));

   mRegs[REG_CTRL1] = REGF_CTRL1_BGON | REGF_CTRL1_OBJON;
   mStrikeCycle = (uint32) -1;
   mStrike = false;

   // set the base palette.
   for (int i = 0; i < palette::NUM_PALETTE_COLORS; i++)
      ((NES *) mpMachine)->setColor(i, mColors[i].r, mColors[i].g, mColors[i].b);
}

// TODO: bleh!
void ppu::setPalette(int index, uint8 value)
{
   value &= 0x3f; // clamp to 64

   ((NES *) mpMachine)->setColor(index, mColors[value].r, mColors[value].g, 
                                 mColors[value].b, value);
}


void ppu::setStrike(int pixel)
{
   mStrikeCycle = nes6502_getcycles(false) + pixel / 3;
   mStrike = true;
}


uint8 ppu::dispatchRead(void *userdata, uint32 offset)
{
   return ((ppu *) userdata)->read(offset);
}

void ppu::dispatchWrite(void *userdata, uint32 offset, uint8 value)
{
   ((ppu *) userdata)->write(offset, value);
}

void ppu::dispatchWriteDMA(void *userdata, uint32 offset, uint8 value)
{
   ((ppu *) userdata)->writeDMA(offset, value);
}

// definite candidate for excommunication.
void ppu::writeDMA(uint32 /*offset*/, uint8 value)
{
   mpOAM->dma(mRegs[REG_OAMADDR], value);
}

uint8 ppu::read(uint32 offset)
{
   uint8 value;

   switch (offset & 7)
   {
   case REG_STAT:
      // only top 3 bits readable
      value = (mRegs[REG_STAT] & 0xe0) | (mLatch & 0x1f);

      if (mStrike && nes6502_getcycles(false) >= mStrikeCycle)
         value |= REGF_STAT_STRIKE;

      // acknowledge vbl interrupt and reset flip flop
      mRegs[REG_STAT] &= ~REGF_STAT_VBLINT;
      mFlipFlop = false;
      break;

   case REG_VDATA:
      // buffered vram reads
      value = mVDataLatch;
      mLatch = mVDataLatch;

      if (vramInaccessible())
      {
         mVDataLatch = 0xff;

         //fprintf(stderr, "inaccessible vram read at $%04x, scanline %d\n", 
         //        mVAddr, mScanline);
      }
      else
      {
         mVDataLatch = PPU_ACCESS(mVAddr);
      }

      mVAddr += (mRegs[REG_CTRL0] & REGF_CTRL0_ADDRINC) ? 32 : 1;
      mVAddr &= 0x3fff;

      break;

   case REG_OAMDATA:
   case REG_CTRL0:
   case REG_CTRL1:
   case REG_OAMADDR:
   case REG_SCROLL:
   case REG_VADDR:
   default:

      value = mLatch;
      break;
   }

   return value;
}

void ppu::write(uint32 offset, uint8 value)
{
   // writes are always latched
   mLatch = value;

   switch (offset & 7)
   {
   case REG_CTRL0:
      mRegs[REG_CTRL0] = value;
      mVAddrLatch &= ~0x0c00;
      mVAddrLatch |= ((value & 3) << 10);
      break;

   case REG_CTRL1:
      mRegs[REG_CTRL1] = value;
      break;

   case REG_STAT:
      // read-only
      break;

   case REG_OAMADDR:
      mRegs[REG_OAMADDR] = value;
      break;

   case REG_OAMDATA:
      mpOAM->write(mRegs[REG_OAMADDR], value);
      mRegs[REG_OAMADDR]++;
      break;

   case REG_SCROLL:
      if (!mFlipFlop)
      {
         mTileXOfs = value & 7;
         mVAddrLatch &= ~0x001f;
         mVAddrLatch |= (value >> 3);
      }
      else
      {
         mVAddrLatch &= ~0x73e0;
         mVAddrLatch |= ((value & 0xf8) << 2);
         mVAddrLatch |= ((value & 7) << 12);
      }

      mFlipFlop ^= true;
      break;

   case REG_VADDR:
      if (!mFlipFlop)
      {
         mVAddrLatch &= ~0xff00;
         mVAddrLatch |= ((value & 0x3f) << 8);
      }
      else
      {
         mVAddrLatch &= ~0x00ff;
         mVAddrLatch |= value;
         mVAddr = mVAddrLatch;
      }

      mFlipFlop ^= true;
      break;

   case REG_VDATA:
      if (mVAddr < 0x3f00)
      {
         // only allow writes if not protected
         if (false == mProtect[mVAddr / PPU_BANKSIZE])
         {            
            if (vramInaccessible())
            {
               //fprintf(stderr, "inaccessible vram write at $%04x, scanline %d\n", 
               //        mVAddr, mScanline);
               value = 0xff; // corrupt
            }

#ifndef NO_TILECACHE
            if (mVAddr < 0x2000 && PPU_ACCESS(mVAddr) != value)
            {
               ASSERT((mVAddr / BYTES_PER_TILE) < 512);
               mTileDirty[mVAddr / BYTES_PER_TILE] = true;
               mAnyTilesDirty = true;
            }
#endif // !NO_TILECACHE

            PPU_ACCESS(mVAddr) = value;
         }
      }
      else
      {
         if (0 == (mVAddr & 15))
         {
            int i;

            for (i = 0; i < 8; i++)
               setPalette(i * 4, value);
         }
         else if (mVAddr & 3)
         {
            setPalette(mVAddr & 0x1f, value);
         }
      }
      
      mVAddr += (mRegs[REG_CTRL0] & REGF_CTRL0_ADDRINC) ? 32 : 1;
      mVAddr &= 0x3fff;

      break;

   default:
      break;
   }
}

// pages 8-11 are the nametable pages, but pages
// 12-15 are mirrors, so always keep them in sync
void ppu::syncHighPages()
{
   mpPage[12] = mpPage[8] - (4 * PPU_BANKSIZE);
   mpPage[13] = mpPage[9] - (4 * PPU_BANKSIZE);
   mpPage[14] = mpPage[10] - (4 * PPU_BANKSIZE);
   mpPage[15] = mpPage[11] - (4 * PPU_BANKSIZE);
}

void ppu::mirror(int nt0, int nt1, int nt2, int nt3)
{
   mpPage[8] = mNameTable + (nt0 * PPU_BANKSIZE) - (8 * PPU_BANKSIZE);
   mpPage[9] = mNameTable + (nt1 * PPU_BANKSIZE) - (9 * PPU_BANKSIZE);
   mpPage[10] = mNameTable + (nt2 * PPU_BANKSIZE) - (10 * PPU_BANKSIZE);
   mpPage[11] = mNameTable + (nt3 * PPU_BANKSIZE) - (11 * PPU_BANKSIZE);

   syncHighPages();
}


// switch in a bank into the ppu, dirtying the entire page
void ppu::bank(int size, int page, uint8 *pMemory, bool writeProtect)
{
   while (size--)
   {
#ifndef NO_TILECACHE
      if (mpPage[page] != pMemory)
      {
         ASSERT(page < PPU_NUMBANKS);
         mPageDirty[page] = true;
         mAnyPagesDirty = true;
      }
#endif // !NO_TILECACHE

      mpPage[page] = pMemory;
      mProtect[page] = writeProtect;
      page++;
   }

   syncHighPages();
}


uint8 *ppu::scanline()
{
   uint8 *pRenderBuf = NULL;

   if (mScanline < 240)
   {
      beginScanline();
      pRenderBuf = renderScanline();
      endScanline();

      cacheOAM();
   }
   else if (mScanline == 241)
   {
      mRegs[REG_STAT] |= REGF_STAT_VBLINT;
   }
   else if (mScanline == 261)
   {
      mRegs[REG_STAT] &= ~REGF_STAT_VBLINT; // does this go down?

      mStrikeCycle = (uint32) -1;
      mStrike = false;

      // special case: garbage rendered scanline with sole
      // intent of caching sprite data for next scanline
      //beginScanline();
      //endScanline();

      cacheOAM();
   }

   mScanline++;
   if (mScanline >= 262)
      mScanline -= 262;

   return pRenderBuf;
}


uint8 *ppu::renderScanline()
{
#ifndef NO_TILECACHE
   updateTileCache();
#endif

   renderBGLine(mScanlineBuffer);
   renderOAMLine(mScanlineBuffer);

   return mScanlineBuffer + 8;
}


void ppu::beginScanline()
{
   if (enabled())
   {
      if (0 == mScanline)
      {
         // transfer whole latch
         mVAddr = mVAddrLatch;
      }
      else
      {
         // reload scanline params
         mVAddr = (mVAddr & ~0x041f) | (mVAddrLatch & 0x041f);
      }
   }
}

void ppu::endScanline()
{
   if (enabled())
   {
      if (7 == (mVAddr >> 12))   // fine y offset rolling over?
      {
         int yTile = (mVAddr >> 5) & 0x1f;
         mVAddr &= ~0x7000;

         if (29 == yTile)
         {
            mVAddr &= ~0x03e0;   // clear y tile
            mVAddr ^= 0x0800;    // toggle hi nametable bit
         }
         else if (31 == yTile)
         {
            mVAddr &= ~0x03e0;   // clear y tile
         }
         else
         {
            mVAddr += 0x20;      // inc y tile
         }
      }
      else
      {
         mVAddr += 0x1000;       // inc y offset
      }
   }
}

#ifndef NO_TILECACHE
#ifdef ASMCODE
extern "C" void draw_cached_bgtile(uint8 *buf, uint8 *src);
#else
static void draw_cached_bgtile(uint8 *buf, uint8 *src)
{
   *buf++ = *src++;
   *buf++ = *src++;
   *buf++ = *src++;
   *buf++ = *src++;
   *buf++ = *src++;
   *buf++ = *src++;
   *buf++ = *src++;
   *buf = *src;
}
#endif
#else /* NO_TILECACHE */

static void draw_bgtile(uint8 *buf, uint8 pat0, uint8 pat1, uint8 colRow)
{
   uint32 pattern = ((pat1 & 0xaa) << 8) | ((pat1 & 0x55) << 1)
                    | ((pat0 & 0xaa) << 7) | (pat0 & 0x55);

   *buf++ = colRow + ((pattern >> 14) & 3);
   *buf++ = colRow + ((pattern >> 6) & 3);
   *buf++ = colRow + ((pattern >> 12) & 3);
   *buf++ = colRow + ((pattern >> 4) & 3);
   *buf++ = colRow + ((pattern >> 10) & 3);
   *buf++ = colRow + ((pattern >> 2) & 3);
   *buf++ = colRow + ((pattern >> 8) & 3);
   *buf = colRow + (pattern & 3);
}
#endif /* NO_TILECACHE */

void ppu::renderBGLine(uint8 *buffer)
{
   uint8 *buf_ptr, *tile_ptr, *attrib_ptr;
   uint32 refresh_vaddr, attrib_base;
   int x_tile, y_tile, tile_count, attrib_shift;

#ifndef NO_TILECACHE
   int tile_base, line_offset, col_base;
#else
   uint32 bg_offset;
   uint8 col_row;
#endif
   uint8 attrib;

   if (0 == (mRegs[REG_CTRL1] & REGF_CTRL1_BGON))
   {
      memset(buffer + 8, 0, 256);
      return;
   }

   buf_ptr = buffer + (8 - mTileXOfs);
   refresh_vaddr = 0x2000 + (mVAddr & 0x0fe0);
   y_tile = (mVAddr >> 5) & 0x1f;
   x_tile = mVAddr & 0x1f;

   tile_ptr = &PPU_ACCESS(refresh_vaddr + x_tile);
   attrib_base = (refresh_vaddr & 0x2c00) + 0x3c0 + ((y_tile & 0x1c) << 1);
   attrib_ptr = &PPU_ACCESS(attrib_base + (x_tile >> 2));
   attrib_shift = (x_tile & 2) + ((y_tile & 2) << 1);
   attrib = *attrib_ptr++;

#ifndef NO_TILECACHE
   col_base = ((attrib >> attrib_shift) & 3) * 512;
   tile_base = (mRegs[REG_CTRL0] & REGF_CTRL0_BGADDR) ? 256 : 0;
   line_offset = (mVAddr >> 12) & 7;
#else
   col_row = ((attrib >> attrib_shift) & 3) << 2;
   bg_offset = (mRegs[REG_CTRL0] & REGF_CTRL0_BGADDR) ? 0x1000 : 0;
   bg_offset += ((mVAddr >> 12) & 7);
#endif

   tile_count = 33;
   while (tile_count--)
   {
      uint8 tile_index = *tile_ptr++;

      // TODO: handle ppu latching (punchout)
#ifndef NO_TILECACHE
      ASSERT(col_base + tile_base + tile_index < 512 * 4);
      ASSERT(line_offset < 8);
      draw_cached_bgtile(buf_ptr, mTileCache[col_base + tile_base + tile_index][line_offset]);
#else
      uint8 *data_ptr = &PPU_ACCESS(bg_offset + tile_index * 16);
      draw_bgtile(buf_ptr, data_ptr[0], data_ptr[8], col_row);
#endif
      buf_ptr += 8;

      x_tile++;
      if (0 == (x_tile & 1))
      {
         if (0 == (x_tile & 3))
         {
            if (32 == x_tile)
            {
               x_tile = 0;
               refresh_vaddr ^= 0x0400; /* flip low nametable bit*/
               attrib_base ^= 0x0400;

               tile_ptr = &PPU_ACCESS(refresh_vaddr);
               attrib_ptr = &PPU_ACCESS(attrib_base);
            }

            attrib = *attrib_ptr++;
         }

         attrib_shift ^= 2;
#ifndef NO_TILECACHE
         col_base = ((attrib >> attrib_shift) & 3) * 512;
#else
         col_row = ((attrib >> attrib_shift) & 3) << 2;
#endif
      }
   }

   /* need to blank left column? */
   if (0 == (mRegs[REG_CTRL1] & REGF_CTRL1_BGNOCLIP))
   {
      uint32 *buf = (uint32 *) (buffer + 8);
      ((uint32 *) buf)[0] = 0;
      ((uint32 *) buf)[1] = 0;
   }
}

#ifndef NO_TILECACHE
void ppu::cacheTile(int tile)
{
   ASSERT(tile < 512);
   uint8 *data_ptr = &PPU_ACCESS(tile * 16);

   for (int line = 0; line < 8; line++)
   {
      uint8 pat0 = data_ptr[0];
      uint8 pat1 = data_ptr[8];
      uint32 pattern = ((pat1 & 0xaa) << 8) | ((pat1 & 0x55) << 1)
                       | ((pat0 & 0xaa) << 7) | (pat0 & 0x55);

      /* cache all 4 possible palette rows */
      for (int row = 0; row < 4; row++)
      {
         int colors = row * 4;
         uint8 *ptr = mTileCache[tile + row * 512][line];

         *ptr++ = colors + ((pattern >> 14) & 3);
         *ptr++ = colors + ((pattern >> 6) & 3);
         *ptr++ = colors + ((pattern >> 12) & 3);
         *ptr++ = colors + ((pattern >> 4) & 3);
         *ptr++ = colors + ((pattern >> 10) & 3);
         *ptr++ = colors + ((pattern >> 2) & 3);
         *ptr++ = colors + ((pattern >> 8) & 3);
         *ptr++ = colors + (pattern & 3);
      }

      data_ptr++;
   }
}


void ppu::cachePage(int page)
{
   // 8 pages of 64 tiles each
   for (int tile = 0; tile < 64; tile++)
      cacheTile(page * 64 + tile);
}


void ppu::updateTileCache()
{
   // catch CHR-ROM paging
   // TODO: build a cache of recently-used swapped-in pages
   if (mAnyPagesDirty)
   {
      mAnyPagesDirty = false;
      for (int page = 0; page < 8; page++)
      {
         if (mPageDirty[page])
         {
            mPageDirty[page] = false;
            cachePage(page);
         }
      }
   }

   if (mAnyTilesDirty)
   {
      mAnyTilesDirty = false;
      // 512 total tiles
      for (int tile = 0; tile < 512; tile++)
      {
         if (mTileDirty[tile])
         {
            mTileDirty[tile] = false;
            cacheTile(tile);
         }
      }
   }
}
#endif // !NO_TILECACHE


int ppu::drawOAMTile(uint8 *pBuf, uint8 *pPriority, uint8 attrib, uint8 pat0,
                     uint8 pat1, uint8 colRow, bool checkStrike)
{
#define DRAW_BEHIND_PIX(pix) \
if (pixels[pix]) \
{ \
   if (0 == (pBuf[pix] & 3)) \
      pBuf[pix] = colRow + pixels[pix]; \
   pPriority[pix] = 1; \
}

#define DRAW_IN_FRONT_PIX(pix) \
if (pixels[pix] && 0 == pPriority[pix]) \
{ \
   pBuf[pix] = colRow + pixels[pix]; \
   pPriority[pix] = 1; \
}

   int pixel = -1;
   uint32 pattern = ((pat1 & 0xaa) << 8) | ((pat1 & 0x55) << 1)
                    | ((pat0 & 0xaa) << 7) | (pat0 & 0x55);

   if (pattern) // not 100% transparent
   {
      uint8 pixels[8];

      if (0 == (attrib & OAMF_HFLIP))
      {
         pixels[0] = (pattern >> 14) & 3;
         pixels[1] = (pattern >> 6) & 3;
         pixels[2] = (pattern >> 12) & 3;
         pixels[3] = (pattern >> 4) & 3;
         pixels[4] = (pattern >> 10) & 3;
         pixels[5] = (pattern >> 2) & 3;
         pixels[6] = (pattern >> 8) & 3;
         pixels[7] = pattern & 3;
      }
      else
      {
         pixels[7] = (pattern >> 14) & 3;
         pixels[6] = (pattern >> 6) & 3;
         pixels[5] = (pattern >> 12) & 3;
         pixels[4] = (pattern >> 4) & 3;
         pixels[3] = (pattern >> 10) & 3;
         pixels[2] = (pattern >> 2) & 3;
         pixels[1] = (pattern >> 8) & 3;
         pixels[0] = pattern & 3;
      }

      if (checkStrike)
      {
         if (pixels[0] && (pBuf[0] & 3))
            pixel = 0;
         else if (pixels[1] && (pBuf[1] & 3))
            pixel = 1;
         else if (pixels[2] && (pBuf[2] & 3))
            pixel = 2;
         else if (pixels[3] && (pBuf[3] & 3))
            pixel = 3;
         else if (pixels[4] && (pBuf[4] & 3))
            pixel = 4;
         else if (pixels[5] && (pBuf[5] & 3))
            pixel = 5;
         else if (pixels[6] && (pBuf[6] & 3))
            pixel = 6;
         else if (pixels[7] && (pBuf[7] & 3))
            pixel = 7;
      }

      if (attrib & OAMF_BEHIND)
      {
         DRAW_BEHIND_PIX(0)
         DRAW_BEHIND_PIX(1)
         DRAW_BEHIND_PIX(2)
         DRAW_BEHIND_PIX(3)
         DRAW_BEHIND_PIX(4)
         DRAW_BEHIND_PIX(5)
         DRAW_BEHIND_PIX(6)
         DRAW_BEHIND_PIX(7)
      }
      else
      {
         DRAW_IN_FRONT_PIX(0)
         DRAW_IN_FRONT_PIX(1)
         DRAW_IN_FRONT_PIX(2)
         DRAW_IN_FRONT_PIX(3)
         DRAW_IN_FRONT_PIX(4)
         DRAW_IN_FRONT_PIX(5)
         DRAW_IN_FRONT_PIX(6)
         DRAW_IN_FRONT_PIX(7)
      }
   }

   return pixel;
}

// evaluate whether or not sprites are in range for the
// next upcoming scanline 
void ppu::cacheOAM()
{
   uint32 vram_base = (mRegs[REG_CTRL0] & REGF_CTRL0_OBJADDR) ? 0x1000 : 0;
   int obj_height = (mRegs[REG_CTRL0] & REGF_CTRL0_OBJ16) ? 16 : 8;

   mRegs[REG_STAT] &= ~REGF_STAT_MAXSPRITE;

   mpOAM->resetCache();
   for (int count = 0; count < MAX_OBJ; count++)
   {
      const obj_t *pObj = mpOAM->getOBJ(count);

      // TODO: cleaner range check?
      if (pObj->y < 240 
          && pObj->y <= mScanline 
          && pObj->y + obj_height > mScanline)
      {
         uint32 vram_addr;

         oamtemp_t temp;
         temp.attrib = pObj->attrib;
         if (count == 0)
            temp.attrib |= OAMF_SPRITE_0;
         else
            temp.attrib &= ~OAMF_SPRITE_0;
         temp.x = pObj->x;

         int offset = mScanline - pObj->y;

         if (16 == obj_height)
         {
            if (offset > 7)
               offset += 8;

            if (pObj->attrib & OAMF_VFLIP)
               offset ^= 23;

            vram_addr = ((pObj->tile & 1) << 12) | ((pObj->tile & 0xfe) << 4);
         }
         else
         {
            if (pObj->attrib & OAMF_VFLIP)
               offset ^= 7;

            vram_addr = vram_base + pObj->tile * 16;
         }

         uint8 *data_ptr = &PPU_ACCESS(vram_addr) + offset;

         temp.pat0 = data_ptr[0];
         temp.pat1 = data_ptr[8];

         if (mpOAM->cache(&temp))
         {
            // sprite limit has been reached.
            mRegs[REG_STAT] |= REGF_STAT_MAXSPRITE;
            break;
         }
      }
   }
}

// actually draw the oam, based on previous line's cache
void ppu::renderOAMLine(uint8 *buffer)
{
   uint8 priority_buf[8 + 256 + 8];
   int current;
   uint32 savecol[2];
   int stat_strikeflag;

   if (0 == (mRegs[REG_CTRL1] & REGF_CTRL1_OBJON))
      return;

   if (0 == (mRegs[REG_CTRL1] & REGF_CTRL1_OBJNOCLIP))
   {
      savecol[0] = ((uint32 *) (buffer + 8))[0];
      savecol[1] = ((uint32 *) (buffer + 8))[1];
   }

   stat_strikeflag = mRegs[REG_STAT] & REGF_STAT_STRIKE;

   memset(priority_buf, 0, sizeof priority_buf);

   for (current = 0; current < mpOAM->getNumCached(); current++)
   {
      const oamtemp_t *pTemp = mpOAM->getCached(current);

      // TODO: vram latching a la punchout
      bool checkStrike = ((pTemp->attrib & OAMF_SPRITE_0) && 0 == stat_strikeflag);
      int strike_pixel = drawOAMTile(buffer + 8 + pTemp->x, 
                                     priority_buf + 8 + pTemp->x,
                                     pTemp->attrib, pTemp->pat0, pTemp->pat1, 
                                     (pTemp->attrib & 3) * 4 + 16,
                                     checkStrike);

      if (checkStrike && -1 != strike_pixel)
         setStrike(pTemp->x + strike_pixel);
   }

   if (0 == (mRegs[REG_CTRL1] & REGF_CTRL1_OBJNOCLIP))
   {
      ((uint32 *) (buffer + 8))[0] = savecol[0];
      ((uint32 *) (buffer + 8))[1] = savecol[1];
   }
}
